/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

#include "class.h"

#include "integer.h"
#include "primitives.h"
#include "garbagecollection.h"
#include "freelist.h"

MethodRef Class_lookupMethod(ClassRef class, ObjectRef messageSymbol) {
  if (class % 2 == 0) {
    NXT_error("ups");
  }
  ObjectRef methodDictionaryRef = Class_get(class)->methodDictionary;
  //STObject* methodDictionary = ObjectMemory_get(methodDictionaryRef);
  U16 methodCount = Object_getSize(methodDictionaryRef) / 2;
  ObjectRef* entry = IndexableObject_get(methodDictionaryRef)->instanceVariables;
  while (methodCount-- > 0) {
    if (*entry++ == messageSymbol) {
      return *entry;
    }
    entry++;
  }
  ClassRef superclass = Class_get(class)->superclass;
  if (superclass != specialObjects->nil) {
    return Class_lookupMethod(superclass, messageSymbol);
  }
  return specialObjects->nil;
}

ObjectRef Class_new(ObjectRef class) {
  U16 instanceVariables = Class_getInstVarCount(Class_get(class));
  return Class_newWithSize(class, instanceVariables);
}

ObjectRef Class_newWithSize(ObjectRef class, U16 sizeRequested) {
  Class* classPtr = Class_get(class);

  NXT_debugString("Allocating object with ");
  NXT_debugNumber(sizeRequested);
  NXT_debugString(" slots...\n");
  int elementSize = Class_getElementSize(classPtr);
  int fullSize = sizeRequested * elementSize + 2;
  if (Class_isVariable(classPtr)) {
    fullSize += 2;
  }

  // Search a free table entry
  Object* object = NULL;
  ObjectRef objectReference;
  for (int i = 0; i < 2; i++) {
    objectReference = ObjectTable_findFreeIndex();
    if (objectReference != 0) {
      // if we get one, try to allocate new memory at the end
      object = Image_allocateRAM(fullSize);
      if (object == NULL) {
        // Nothing found, search freelist
        objectReference = Freelist_allocate(fullSize);
        if (objectReference == 0) {
          // still nothing found, do full collect
          GarbageCollection_collect();
          continue;
        }
        // get real pointer
        object = ObjectTable_getAddress(objectReference);
      } else {
        // we could allocate some memory, use it
        ObjectTable_setAddress(objectReference, object);
      }
    } else {
      // we did not get a free table entry anymore, use freelist
      objectReference = Freelist_allocate(fullSize);
      if (objectReference == 0) {
        // still nothing found, now do full collect
        GarbageCollection_collect();
        continue;
      }
      // get real pointer
      object = ObjectTable_getAddress(objectReference);
    }
    if (object) {
      break;
    }
  }
  if (object == NULL) {
    NXT_error("Memory full");
  }

  // Reestablish class pointer due to gc
  classPtr = Class_get(class);
  object->klass = class;
  ObjectRef* instanceVariables;
  if (Class_isVariable(classPtr)) {
    ((IndexableObject*)object)->size = sizeRequested;
    instanceVariables = ((IndexableObject*)object)->instanceVariables;
  } else {
    instanceVariables = object->instanceVariables;
  }
  if (elementSize == 2) {
    ObjectTable_setReferencingInstances(objectReference, TRUE);
    for (int i = 0; i < sizeRequested; i++) {
      instanceVariables[i] = specialObjects->nil;
    }
  } else {
    ObjectTable_setReferencingInstances(objectReference, FALSE);
    for (int i = 0; i < (sizeRequested + 1) / 2; i++) {
      instanceVariables[i] = 0;
    }
  }

  objectCount++;
  ObjectTable_setFree(objectReference, FALSE);
  ObjectTable_setReferenceCount(objectReference, 0);
  ObjectTable_setRenewed(objectReference, FALSE);
  ObjectTable_setMarked(objectReference, FALSE);
  ObjectTable_setIndexable(objectReference, Class_isVariable(classPtr));
  return objectReference;
}

void Class_primBasicNew(PrimitiveFrame* frame) {
  Primitive_return(frame, Class_new(frame->self));
}

void Class_primBasicNewSize(PrimitiveFrame* frame) {
  U16 sizeRequested = SmallInteger_fromReference(Primitive_argumentAt(frame, 0));
  Primitive_return(frame, Class_newWithSize(frame->self, sizeRequested));
}

void Class_primFirstInstance(PrimitiveFrame* frame) {
  Primitive_return(frame, Class_findFirstInstance(frame->self));
}
