/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

/** \file class.h
 *  \brief Functions for classes.
 */

#ifndef _CLASS_H_
#define _CLASS_H_

#include "object.h"
#include "method.h"
#include "string.h"
#include "integer.h"
#include "image.h"

/** \brief Type redefinition to indicate Class object references. */
typedef ObjectRef ClassRef;

/** \brief Class representation in memory.
 *
 * This structure describes a class object within the memory. NXTalk needs
 * it to lookup methods and to retrieve class information like the size of
 * its instances.
 */
struct Class {

  /** \brief Class reference. */
  ClassRef klass;

  /** \brief The method dictionary of this class.
   *
   * At the moment, the method dictionary of a class is a simple array, which
   * contains (selector, method) pairs. In order to find a method, a
   * sequential search is necessary.
   */
  ObjectRef methodDictionary;

  /** \brief Reference to the superclass of this class. */
  ClassRef superclass;

  /** \brief Name of this class. */
  SymbolRef name;

  /** \brief Class format.
   *
   * Encodes class structure bits in a tagged-integer. The two lowest
   * bits represent the element size. E.g. 1 for bytes and 2 for half-words
   * or regular instances. Next, 5 bits encode the number of variables
   * instances of this class have. Finally, the 8-th bit indicates, whether
   * instances of this class are indexable or not. Several convenience
   * functions to access these values exist.
   */
  SmallInteger format;

  /** \brief Names of the instance variables.
   *
   * This variable references an array of the names of the instance
   * variables.
   */
  ObjectRef instVarNames;
} __attribute__ ((packed));
typedef struct Class Class;

/** \brief Resolve a Class reference.
 *
 * Retrieves a direct pointer to a Class object.
 *
 * \param classRef The reference to the class.
 * \return The direct pointer to the Class object in memory.
 */
static inline Class* Class_get(ClassRef classRef) {
   return (Class*)ObjectTable_getAddress(classRef);
}

/** \brief Resolve a Class reference for writing.
 *
 * Retrieves a direct pointer to a Class object ensuring it lies entirely
 * within RAM.
 *
 * \param classRef The reference to the class.
 * \return The direct pointer to the Class object in RAM.
 */
static inline Class* Class_getForWrite(ClassRef classRef) {
  return (Class*)Object_getForWrite(classRef);
}

/** \brief Looks up a method.
 *
 * Finds the method coressponding to the given message name in the method
 * dictionaries of the class or its superclasses.
 *
 * \param class         The class where the search should be started.
 * \param messageSymbol Name of the method to search.
 * \return Found method or nil, if none was found.
 */
MethodRef Class_lookupMethod(ClassRef class, ObjectRef messageSymbol);

/** \brief Finds the first instance in memory.
 *
 * Finds the first instance of the given class in the memory. In fact
 * the instance with the lowest reference number is found.
 *
 * \param class A reference to the class.
 * \return A reference to the class' first instance.
 */
static inline ObjectRef Class_findFirstInstance(ObjectRef class) {
  for (int reference = minReference; reference <= MAX_REFERENCE; reference += 2) {
    if (!ObjectTable_isFree(reference) && Object_get(reference)->klass == class) {
      return reference;
    }
  }
  return specialObjects->nil;
}

/** \brief Creates a new instance of the given class.
 *
 * \param klass A reference to the class.
 * \return The new instance.
 */
ObjectRef Class_new(ClassRef klass);

/** \brief Creates a new instance of the given class with the given size.
 *
 * \param klass         A reference to the class.
 * \param sizeRequested The requested number of data slots for the new
 *                      instance.
 * \return The new instance.
 */
ObjectRef Class_newWithSize(ClassRef klass, U16 sizeRequested);

/** \name Getters
 * @{
 */

/** \brief Fetches the class format bitfield.
 *
 * \param self The class, which format shall be fetched.
 * \return A native integer containing the format bitfield.
 */
static inline U16 Class_getFormat(const Class* self) {
  return SmallInteger_fromReference(self->format);
}

/** \brief Returns the number of instance variables for this class.
 *
 * \param self A direct pointer to the class.
 * \return Number of instance variables.
 */
static inline U16 Class_getInstVarCount(const Class* self) {
  return (Class_getFormat(self) >> 2) & 0x1F;
}

/** \brief Indicates, whether instances of this class are indexable.
 *
 * \param self A direct pointer to the class.
 * \return TRUE, if instances are indexable, FALSE if not.
 */
static inline bool Class_isVariable(const Class* self) {
  return (Class_getFormat(self) >> 7) & 0x01;
}

/** \brief Returns the element size the instances of the given class.
 *
 * \param self A direct pointer to the class.
 * \return  1 for byte-oriented instances. 2 for half-word-oriented or regular
 *          instances.
 */
static inline U16 Class_getElementSize(const Class* self) {
  return Class_getFormat(self) & 0x03;
}
/** @} */

/** \name Primitives
 * @{
 */

/** \brief Create a new instance of this (non-indexable class).
 *
 * \param frame Contains necessary context information for the primitive.
 */
void Class_primBasicNew(PrimitiveFrame* frame);

/** \brief Create a new instance of this indexable class with the given size.
 *
 * \param frame Contains necessary context information for the primitive.
 */
void Class_primBasicNewSize(PrimitiveFrame* frame);

/** \brief Find first instance of this class in memory.
 *
 * \param frame Contains necessary context information for the primitive.
 */
void Class_primFirstInstance(PrimitiveFrame* frame);
/**
 * @}
 */

#endif // _CLASS_H_
