/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

/** \file frame.h
 * \author Martin Beck
 * \brief Types and functions for event handling.
 *
 * Event handling for the sensors, buttons and motors does not support
 * interrupts. Hence, the processes waiting for events can specify type
 * and event condition. These are checked then by polling the necessary
 * data sources.
 */

#ifndef _EVENT_H_
#define _EVENT_H_

#include "object.h"
#include "integer.h"
#include "class.h"

/** \brief Type redefinition to indicate Event objects. */
typedef ObjectRef EventRef;

/** \brief Memory structure of an even description.
 *
 * Describes the event within Smalltalk, which is expected by the
 * waiting process in order to continue execution. In order to save
 * classes and thus space, the type of the event, the value and the
 * condition are expressed by constant numbers.
 */
struct Event {
  /** \brief Class refrence. */
  ClassRef klass;

  /** \brief An event can be in a linked list. */
  EventRef next;

  /** \brief Integer representing the event type. */
  SmallInteger type;

  /** \brief Integer representing an expected value. */
  Integer value;

  /** \brief SmallInteger representing the condition to come true. */
  SmallInteger condition;
};
typedef struct Event Event;

/** \brief Resolves an event reference to a direct pointer.
 *
 * \param reference The reference to the event to resolve.
 * \return Direct pointer to the resolved event.
 */
static inline Event* Event_get(EventRef reference) {
  return (Event*)Object_get(reference);
}

/** \brief Enumeration of the different event types. */
enum EventType {
  /** \brief Certain system time reached. */
  TIME,

  /** \brief Certain button pressed. */
  BUTTON,

  /** \brief Certain value reached for analog sensor 1. */
  SENSOR_1,

  /** \brief Certain value reached for analog sensor 2. */
  SENSOR_2,

  /** \brief Certain value reached for analog sensor 3. */
  SENSOR_3,

  /** \brief Certain value reached for analog sensor 4. */
  SENSOR_4,

  /** \brief Certain value reached for the rotation sensor of motor A. */
  MOTOR_A,

  /** \brief Certain value reached for the rotation sensor of motor B. */
  MOTOR_B,

  /** \brief Certain value reached for the rotation sensor of motor C. */
  MOTOR_C
};

/** \brief Enumeration of conditions for numerical values. */
enum EventCondition {
  /** \brief The read value has to be greater than the specified. */
  GREATER_THAN,

  /** \brief The read value has to be less than the specified. */
  LESS_THAN
};

/** \brief Checks, whether the given event description is currently present.
 *
 * \param event Event description, which shall checked whether the described
 *              conditions are currently present.
 * \return TRUE, if the event is currently present. FALSE, if not.
 */
bool Event_check(Event* event);

#endif // _EVENT_H_
