/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

#include "frame.h"

#include "image.h"
#include "interpreter.h"
#include "primitives.h"
#include "scheduler.h"

static FrameRef framePool[MAX_FRAME_SIZE + 1];
static int framePoolSize = 0;

static FrameRef Frame_fromPool(U16 frameSize) {
  for (int i = frameSize; i <= MAX_FRAME_SIZE; i++) {
    if (framePool[i] != 0) {
      FrameRef frame = framePool[i];
      framePool[i] = Frame_get(frame)->sender;
      framePoolSize--;
      ObjectTable_setReferenceCount(frame, 0);
      return frame;
    }
  }
  if (frameSize > MAX_FRAME_SIZE) {
    NXT_error("Too large frame");
  }
  return 0;
}

FrameRef* Frame_getPool() {
  return framePool;
}

static void Frame_toPool(FrameRef frame) {
  // insert it at the head of the empty frame chain
  U16 frameSize = Object_getSize(frame);
  Frame_get(frame)->sender = framePool[frameSize];
  framePool[frameSize] = frame;
  framePoolSize++;
}

static FrameRef Frame_new(ClassRef frameClass, U8 argCount, U8 tempCount, Frame** framePtr) {
  FrameRef frame = Frame_fromPool(FRAME_SIZE + argCount + tempCount);
  if (frame == 0) {
    frame = Class_newWithSize(frameClass, FRAME_SIZE + argCount + tempCount);
  }

  // its safe to not check for ROM object here and use only Frame_get(), because the object is new
  Frame* self = Frame_get(frame);
  *framePtr = self;
  // the method and block classes always have VERY high reference count, thus no explicit RC is needed
  self->klass = frameClass;
  self->sender = specialObjects->nil;
  Frame_setPC(self, 1);
  Frame_setStackPointer(self, 0);
  return frame;
}

FrameRef Frame_newMethod(ObjectRef receiver, ObjectRef method) {
  U8 tempCount = Method_getTempCount(Method_get(method));
  U8 argCount = Method_getArgumentCount(Method_get(method));
  Frame* self;
  FrameRef frame = Frame_new(specialObjects->NXTMethodFrame, argCount, tempCount, &self);
  // No explicit reference count decrement needed
  Object_incReferenceCount(receiver);
  Object_incReferenceCount(method);
  self->shared.method.receiver = receiver;
  self->shared.method.method = method;
  return frame;
}

FrameRef Frame_newBlock(ObjectRef block, FrameRef home) {
  U8 argCount = Method_getArgumentCount(Method_get(block));
  U8 tempCount = Method_getTempCount(Method_get(block));
  Frame* self;
  FrameRef frame = Frame_new(specialObjects->NXTBlockFrame, argCount, tempCount, &self);
  Object_incReferenceCount(block);
  Object_incReferenceCount(home);
  self->shared.block.block = block;
  self->shared.block.home = home;
  return frame;
}

void Frame_free(FrameRef frame) {
  Frame* self = Frame_get(frame);
  Object_decReferenceCount(self->sender);
  self->sender = specialObjects->nil;
  Object_decReferenceCount(self->shared.method.method);
  self->shared.method.method = specialObjects->nil;
  Object_decReferenceCount(self->shared.method.receiver);
  self->shared.method.receiver = specialObjects->nil;
  // nil out args and temps (RC is already decremented)
  for (int i = 0; i < self->size - FRAME_SIZE; i++) {
    Object_decReferenceCount(self->argsAndTemps[i]);
    self->argsAndTemps[i] = specialObjects->nil;
  }
  if (framePoolSize < MAX_FRAME_POOL_SIZE) {
    Frame_toPool(frame);
  } else {
    Object_forceFree(frame);
  }
}

void Frame_restart(FrameRef frame) {
  // clear stack
  Interpreter_stackSetTopIndex(Frame_getStackPointer(Frame_get(frame)));
  // reexecute frame
  Frame_setPC(Frame_get(frame), 1);
}

ObjectRef Frame_getArgument(const Frame* self, U16 index) {
  if (index >= Frame_getArgumentCount(self)) {
    if (Frame_executingBlock(self)) {
      return Frame_getArgument(BlockFrame_getHome(self), index - Frame_getArgumentCount(self));
    } else {
      NXT_error("Argument index out of bounds!\n");
      return 0;
    }
  } else {
    return self->argsAndTemps[index];
  }
}

ObjectRef Frame_getTemporary(const Frame* self, U16 index) {
  if (index >= Frame_getTempCount(self)) {
    if (Frame_executingBlock(self)) {
      return Frame_getTemporary(BlockFrame_getHome(self), index - Frame_getTempCount(self));
    } else {
      NXT_error("Temporary index out of bounds!\n");
      return 0;
    }
  } else {
    return self->argsAndTemps[Frame_getArgumentCount(self) + index];
  }
}

ObjectRef Frame_setTemporary(Frame* self, U16 index, ObjectRef value) {
  if (index >= Frame_getTempCount(self)) {
    // access out of our temporaries
    if (Frame_executingBlock(self)) {
      // we got a BlockFrame, use temporaries from home
      return Frame_setTemporary(Frame_getForWrite(self->shared.block.home), index - Frame_getTempCount(self), value);
    } else {
      // access violation
      NXT_error("Temporary index out of bounds!\n");
      return 0;
    }
  } else {
    // Frame is on stack, calculate temp position on stack
    U16 argCount = Frame_getArgumentCount(self);
    ObjectRef oldValue = self->argsAndTemps[argCount + index];
    self->argsAndTemps[argCount + index] = value;
    return oldValue;
  }
}

void BlockFrame_primValue(PrimitiveFrame* frame) {
  U8 argumentCount = Method_getArgumentCount(Frame_getMethod(Frame_get(frame->self)));
  Interpreter_storeInstructionStream();
  Interpreter_executeFrame(frame->self, argumentCount);
  Interpreter_loadInstructionStream();
}

void BlockFrame_primRestart(PrimitiveFrame* frame) {
  Primitive_clearStack(frame);
  Frame_restart(currentFrame);
  Interpreter_loadInstructionStream();
}
