/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

/** \file frame.h
 *  \brief Functions and primitives for the Frame classes.
 */

#ifndef _FRAME_H_
#define _FRAME_H_

#include "method.h"
#include "class.h"
#include "integer.h"
#include "boolean.h"

/** \brief Type redefinition to indicate Frame object references. */
typedef ObjectRef FrameRef;

/** \brief Frame representation in memory.
 *
 * This structure represents an activation frame context in memory. It can
 * encode both a block context and a method context. Besides necessary
 * activation data it also contains the frame arguments and temporary
 * values.
 */
struct Frame {

  /** \brief Class reference.
   *
   * Can reference either NXTBlockFrame class or NXTMethodFrame class.
   */
  ClassRef klass;

  /** \brief Number of instance variables.
   *
   * Frames have a dynamic size as they also store their arguments and
   * temporaries.
   */
  U16 size;

  /** \brief Sender of this frame.
   *
   * References the frame context, which activated and thus is the
   * sender of this context.
   */
  FrameRef sender;

  /** \brief Program counter.
   *
   * An integer indicating the currently executed bytecode in the
   * corresponding method. It is already encoded as a tagged-integer
   * and can thus be directly accessed from within Smalltalk.
   */
  SmallInteger pc;

  /** \brief Stack pointer.
   *
   * An integer indicating the top of the stack, when this frame
   * was executed. As each executed frame can only belong to exactly
   * one process, this integer implicitly references that process'
   * stack.
   */
  SmallInteger stackPointer;

  /** \brief Mixed data for block and method frame contexts.
   *
   * As method and block frames have the same number of named instance
   * variables and in order to avoid additional structs, a frame contains
   * this shared space in order to access the block context and the method
   * context variables using meaningful names.
   */
  union Shared {

    /** \brief NXTBlockFrame specific data.
     *
     * Contains data fields of the frame which are only important for block
     * frame contexts.
     */
    struct BlockFrame {

      /** \brief Compiled block reference.
       *
       * A reference to the object representing the compiled code of the
       * block executed by this frame.
       */
      BlockRef block;

      /** \brief Home of this block context.
       *
       * A reference to the frame context, in which this block frame
       * was created.
       */
      FrameRef home;
    } __attribute__ ((packed)) block;

    /** \brief NXTMethodFrame specific data.
     *
     * Contains data fields of the frame which are only important for method
     * frame contexts.
     */
    struct MethodFrame {

      /** \brief Compiled method reference.
       *
       * A reference to the object representing the compiled code of the
       * method executed by this frame.
       */
      MethodRef method;

      /** \brief Receiver of method activation.
       *
       * Reference to the receiver object used during this method
       * activation.
       */
      ObjectRef receiver;
    } __attribute__ ((packed)) method;
  } __attribute__ ((packed)) shared;

  /** \brief Arguments and temporaries.
   *
   * Here, first every argument is referenced and afterwards every temporary
   * is stored. On execution of this frame, the arguments are copied from the
   * stack and the temps are filled with nil.
   */
  ObjectRef argsAndTemps[];
} __attribute__ ((packed));
typedef struct Frame Frame;

/** \brief Size of a frame with out arguments and temporaries. */
#define FRAME_SIZE 5

/** \brief Maximum size of a frame including arguments and temporaries. */
#define MAX_FRAME_SIZE 20

/** \brief Maximum size of the frame pool. */
#define MAX_FRAME_POOL_SIZE 100

/** \brief Creates a new method frame.
 *
 * Instanciates a new method frame, which has the correct number of argument
 * and temporary variables as specified by the given compiled method.
 *
 * \param receiver  The receiver object, in which context the method shall be
 *                  executed.
 * \param method    The compiled method, which shall be executed with this
 *                  method frame.
 * \return A reference to the new method frame object.
 */
FrameRef Frame_newMethod(ObjectRef receiver, MethodRef method);

/** \brief Creates a new block frame.
 *
 * Instanciates a new block frame, which has the correct number of argument
 * and temporary variables as specified by the given compiled block.
 *
 * \param block The compiled block, which shall be executed with this block
 *              frame.
 * \param home  The frame, in which this block frame shall search for outer
 *              variables and the receiver.
 * \return A reference to the new block frame object.
 */
FrameRef Frame_newBlock(BlockRef block, FrameRef home);

/** \brief Frees a frame context.
 *
 * Finally frees a frame context, which may mean it is simply inserted
 * into the frame pool or freed like any other object, if the pool is
 * already full.
 *
 * \param frame Frame to be freed.
 */
void Frame_free(FrameRef frame);

/** \brief Restarts a frame context.
 *
 * Restarts a frame context by resetting the program counter and
 * updating the interpreter's stack.
 *
 * \param frame The frame, which shall be restarted. Has to be a sender of
 *              the frame currently executed by the interpreter.
 */
void Frame_restart(FrameRef frame);

/** \brief Fetches an argument of the frame.
 *
 * Fetches an argument of the frame itself or its home frame, if a
 * block frame is specified.
 *
 * \param self  The frame to operate on.
 * \param index The index of the argument to fetch.
 * \return The value of the fetched argument.
 */
ObjectRef Frame_getArgument(const Frame* self, U16 index);

/** \brief Fetches a temporary of the frame.
 *
 * Fetches a temporary of the frame itself or its home frame, if a
 * block frame is specified.
 *
 * \param self  The frame to operate on.
 * \param index The index of the temporary to fetch.
 * \return The value of the fetched temporary.
 */
ObjectRef Frame_getTemporary(const Frame* self, U16 index);

/** \brief Updates a temporary of the frame.
 *
 * Updates a temporary of the frame itself or its home frame, if a
 * block frame is specified.
 *
 * \param self  The frame to operate on.
 * \param index The index of the temporary to fetch.
 * \param value The new value of the temporary.
 * \return The set value of the temporary.
 */
ObjectRef Frame_setTemporary(Frame* self, U16 index, ObjectRef value);

/** \brief Resolves a reference to a Frame.
 *
 * Fetches a direct pointer to a Frame from the object table.
 *
 * \param frameRef A reference to a frame object.
 * \return Direct pointer to the Frame.
 */
static inline Frame* Frame_get(FrameRef frameRef) {
  return (Frame*)ObjectTable_getAddress(frameRef);
}

/** \brief Resolves a reference to a Frame for writing.
 *
 * Fetches a direct pointer to a Frame from the object table and moves the
 * object to the RAM. This is needed if writing operations shall be performed
 * on the frame object.
 *
 * \param frameRef A reference to a frame object.
 * \return Direct pointer to the Frame, which is ensured to live in RAM.
 */
static inline Frame* Frame_getForWrite(FrameRef frameRef) {
  return (Frame*)Object_getForWrite(frameRef);
}

/** \brief Fetches a block frame's home context.
 *
 * \param self The block frame, which home context shall be fetched.
 * \return A direct pointer to the home context for reading.
 */
static inline const Frame* BlockFrame_getHome(const Frame* self) {
  return Frame_get(self->shared.block.home);
}

/** \brief Indicates, whether the given frame is a block frame.
 *
 * \param self The frame, whose type shall be checked.
 * \return TRUE, if the given frame is a block frame, FALSE otherwise.
 */
static inline bool Frame_executingBlock(const Frame* self) {
  return self->klass == specialObjects->NXTBlockFrame;
}

/** \name Getters
 * @{
 */

/** \brief Returns the pool of currently unused frames sorted by their size.
 *
 * \return A pointer to the frame pool. */
FrameRef* Frame_getPool(void);

/** \brief Fetches the home method frame of a block frame.
 *
 * Recursively fetches the method frame, within the given block frame was
 * created.
 *
 * \param self The frame, whose home method frame to compute.
 * \return A reference to the home method frame.
 */
static inline FrameRef Frame_getHomeMethodFrame(const Frame* self) {
  while (Frame_executingBlock(BlockFrame_getHome(self))) {
    self = BlockFrame_getHome(self);
  }
  return self->shared.block.home;
}

/** \brief Fetches a frame's underlying method (or block).
 *
 * \param self The frame, which underlying method shall be fetched.
 * \return  A direct pointer to the compiled Method or Block, this
 *          frame executes.
 */
static inline const Method* Frame_getMethod(const Frame* self) {
  return Method_get(self->shared.method.method);
}

/** \brief Fetches the frame's receiver.
 *
 * Recursively fetches the frame's receiver regardless of its type (e.g.
 * whether it is a method frame or a block frame).
 *
 * \param self The frame, which receiver has to be fetched.
 * \return A reference to the receiver used in the given frame.
 */
static inline ObjectRef Frame_getReceiver(const Frame* self) {
  if (Frame_executingBlock(self)) {
    self = Frame_get(Frame_getHomeMethodFrame(self));
  }
  return self->shared.method.receiver;
}

/** \brief Gets the program counter.
 *
 * \param self The frame, which pc shall be get.
 * \return The program counter converted to a native int.
 */
static inline U16 Frame_getPC(const Frame* self) {
  return SmallInteger_fromReference(self->pc);
}

/** \brief Gets the stack pointer.
 *
 * \param frame The frame, which stack pointer shall be read.
 * \return The stack pointer converted to a native int.
 */
static inline U16 Frame_getStackPointer(const Frame* frame) {
  return SmallInteger_fromReference(frame->stackPointer) - 1;
}

/** \brief Gets the argument count of the underlying method (or block).
 *
 * \param self The frame, for which to calculate the argument count.
 * \return The argument count.
 */
static inline U8 Frame_getArgumentCount(const Frame* self) {
  return Method_getArgumentCount(Method_get(self->shared.method.method));
}

/** \brief Gets the temporary count of the underlying method (or block).
 *
 * \param self The frame, for which to calculate the temporary count.
 * \return The temporary count.
 */
static inline U8 Frame_getTempCount(const Frame* self) {
  return Method_getTempCount(Method_get(self->shared.method.method));
}

/** \brief Retrieves the containing class of this frame's method (or block).
 *
 * \param self The frame.
 * \return  The class, which contains the compiled method or block used
 *          by the given frame.
 */
static inline ObjectRef MethodFrame_getOwner(const Frame* self) {
  if (Frame_executingBlock(self)) {
    self = Frame_get(Frame_getHomeMethodFrame(self));
  }
  return Method_getMethodClass(Frame_getMethod(self));
}
/** @} */


/** \name Setters
 * @{
 */

/** \brief Sets the program counter.
 *
 * \param self  The frame, whose pc shall be set.
 * \param pc    The actual program counter in bytes.
 */
static inline void Frame_setPC(Frame* self, U16 pc) {
  // No ref counting for numbers
  self->pc = SmallInteger_toReference(pc);
}

/** \brief Sets the stack pointer.
 *
 * \param self          The frame, whose stack pointer shall be set.
 * \param stackPointer  The new stack pointer value.
 */
static inline void Frame_setStackPointer(Frame* self, U16 stackPointer) {
  // No ref counting for numbers (and nil)
  self->stackPointer = SmallInteger_toReference(stackPointer + 1);
}
/** @} */

/** \name Primitives
 * @{
 */

/** \brief Executes a block frame.
 *
 * This primitive executes a block and uses as many arguments from the stack
 * as the underlying block expects.
 *
 * \param frame Contains necessary context information for the primitive.
 */
void BlockFrame_primValue(PrimitiveFrame* frame);

/** \brief Restarts a block frame.
 *
 * This primitive restarts a block frame at the beginning of the bytecodes.
 *
 * \param frame Contains necessary context information for the primitive.
 */
void BlockFrame_primRestart(PrimitiveFrame* frame);
/**
 * @}
 */

#endif // _FRAME_H_
