/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

#include "freelist.h"

#define MIN_SIZE 2
#define MAX_SIZE 30
#define ANY_SIZE (MAX_SIZE + 2)
#define FREELIST_COUNT ((MAX_SIZE - MIN_SIZE) / 2 + 1)
#define ANY_SIZE_LIST FREELIST_COUNT

ObjectRef freelists[FREELIST_COUNT + 1];
int freeBlocks;

static ObjectRef Freelist_freelist(int size) {
  if (size >= ANY_SIZE) {
    // return freelist for biggest ones
    return freelists[ANY_SIZE_LIST];
  } else {
    return freelists[(size - 4) / 2];
  }
}

static void Freelist_setFreelist(int size, ObjectRef value) {
  if (size >= ANY_SIZE) {
    freelists[ANY_SIZE_LIST] = value;
  } else {
    freelists[(size - 4) / 2] = value;
  }
}

ObjectRef Freelist_allocate(U16 requestedSize) {
  requestedSize += requestedSize & 1;
  // first, try to find exact match
  if (requestedSize < ANY_SIZE) {
    ObjectRef freeRef = Freelist_freelist(requestedSize);
    if (freeRef != 0) {
      FreeEntry* free = (FreeEntry*)ObjectTable_getAddress(freeRef);
      Freelist_setFreelist(requestedSize, free->next);
      freeBlocks--;
      return freeRef;
    }
  }
  int size = MIN(requestedSize + sizeof(struct FreeEntry), ANY_SIZE);
  // now try other known sizes which could be split up into new entries
  while (size <= MAX_SIZE) {
    ObjectRef freeRef = Freelist_freelist(requestedSize);
    if (freeRef != 0) {
      // split up and insert new free block into its list
      FreeEntry* free = (FreeEntry*)ObjectTable_getAddress(freeRef);
      FreeEntry* newFree = (FreeEntry*)((U8*)free + requestedSize);
      ObjectRef newFreeRef = ObjectTable_findFreeIndex();
      if (newFreeRef == 0) {
        // no free entries in object table left, can't split up
        return 0;
      }
      ObjectTable_setAddress(newFreeRef,  (Object*)newFree);
      ObjectTable_setFree(newFreeRef, TRUE);
      newFree->size = size - requestedSize;
      newFree->next = Freelist_freelist(newFree->size);
      Freelist_setFreelist(newFree->size, newFreeRef);
      return freeRef;
    }
    size += 2;
  }
  // Now search the big free list, containing all other sizes
  FreeEntry* oldFree = NULL;
  ObjectRef freeRef = Freelist_freelist(ANY_SIZE);
  while (freeRef != 0) {
    FreeEntry* free = (FreeEntry*)ObjectTable_getAddress(freeRef);
    if (free->size == requestedSize) {
      // found exact match
      if (oldFree) {
        oldFree->next = free->next;
      } else {
        Freelist_setFreelist(size, free->next);
      }
      freeBlocks--;
      return freeRef;
    } else if (free->size >= requestedSize + sizeof(struct FreeEntry)) {
      // split up and insert new free block into its list
      FreeEntry* newFree = (FreeEntry*)((U8*)free + requestedSize);
      ObjectRef newFreeRef = ObjectTable_findFreeIndex();
      if (newFreeRef == 0) {
        // no free entries in object table left
        return 0;
      }
      ObjectTable_setAddress(newFreeRef, (Object*)newFree);
      ObjectTable_setFree(newFreeRef, TRUE);
      newFree->size = free->size - requestedSize;
      newFree->next = Freelist_freelist(newFree->size);
      Freelist_setFreelist(newFree->size, newFreeRef);
      // remove found entry from its freelist
      if (oldFree) {
        oldFree->next = free->next;
      } else {
        Freelist_setFreelist(size, free->next);
      }
      return freeRef;
    }
    oldFree = free;
    freeRef = free->next;
  }
  // Nothing found, goodbye
  return 0;
}

bool Freelist_isVerySmall(ObjectRef reference) {
  return ObjectTable_isIndexable(reference);
}

static void Freelist_setVerySmall(ObjectRef reference, bool verySmall) {
  ObjectTable_setIndexable(reference, verySmall);
}

void Freelist_free(ObjectRef object, U16 size) {
  FreeEntry* free = (FreeEntry*)ObjectTable_getAddress(object);
  free->next = Freelist_freelist(size);
  Freelist_setFreelist(size, object);
  freeBlocks++;
  if (size == 2) {
    Freelist_setVerySmall(object, TRUE);
  } else {
    Freelist_setVerySmall(object, FALSE);
    free->size = size;
  }
}

void Freelist_initialize() {
  freeBlocks = 0;
  for (int i = 0; i < FREELIST_COUNT + 1; i++) {
    freelists[i] = 0;
  }
}
