/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

/** \file freelist.h
 * \author Martin Beck
 * \brief Functions and variables for managing lists of free ram.
 */

#ifndef _FREELIST_H_
#define _FREELIST_H_

#include "object.h"

/** \brief Structure of a free block in the object heap.
 *
 * Free blocks are stored instead of an object, if it was freed by the
 * reference counting. Its data is replaced with the data specified below
 * describing the size of the block and references the next block in the
 * list.
 */
struct FreeEntry {
  /** \brief Reference to the next object in the current free list. */
  ObjectRef next;

  /** \brief Size of this free block.
   *
   * Only present, if this is no very small block.
   *
   * \see Freelist_isVerySmall
   */
  U16 size;
} __attribute__ ((packed));
typedef struct FreeEntry FreeEntry;

/** \brief Initialize the freelist module. */
void Freelist_initialize(void);

/** \brief Try to allocate memory from the freelist
 *
 * Searches the freelists for block with the given size or larger,
 * splits it if necessary (and possible) and returns its object reference.
 *
 * \param size The size of the contiguous block in bytes.
 * \return The object reference to the found block. 0, if none was found.
 */
ObjectRef Freelist_allocate(U16 size);

/** \brief Frees an object.
 *
 * Frees the memory block of the given object with the given size and
 * inserts it in the corresponding freelist.
 *
 * \param object The object to free.
 * \param size   Size of the object to free in bytes.
 */
void Freelist_free(ObjectRef object, U16 size);

/** \brief Indicates, whether a free block is very small.
 *
 * As the smallest possible object is only 2 bytes large, a whole FreeEntry
 * would occupy 4 bytes with the size. Thus, this function reuses the
 * indexable bit in the object table, which is of course not needed for
 * free blocks, in order to indicate free blocks, which are only 2 bytes
 * large and thus do not have and need a size entry.
 *
 * \param reference The free reference to check whether it is 2 bytes large.
 * \return TRUE, if the free block is 2 bytes large. FALSE, if not.
 */
bool Freelist_isVerySmall(ObjectRef reference);

#endif // _FREELIST_H_
