/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

#include "garbagecollection.h"

#include "frame.h"
#include "interpreter.h"
#include "freelist.h"

// Collection statistics
U32 lastInvocation;
U32 counter;
U32 aggregatedCollectionTime;
U32 aggregatedInbetweenTime;
U32 markedObjects;

void GarbageCollection_resetStatistics() {
  lastInvocation = 0;
  counter = 0;
  aggregatedCollectionTime = 0;
  aggregatedInbetweenTime = 0;
}

static void GarbageCollection_updateStatistics(bool startOfGC) {
  if (startOfGC == TRUE) {
    if (counter > 0) {
      aggregatedInbetweenTime += NXT_getSystemTime() - lastInvocation;
    }
    lastInvocation = NXT_getSystemTime();
    counter++;
  } else {
    if (lastInvocation != 0) {
      aggregatedCollectionTime += NXT_getSystemTime() - lastInvocation;
      lastInvocation = NXT_getSystemTime();
    }
  }
}

void GarbageCollection_printStatistics() {
  NXTDisplay_clear();
  NXT_printString("GC Statistics:\n");
  NXT_printNumber(counter); NXT_printString("\n");
  NXT_printNumber(aggregatedCollectionTime); NXT_printString("\n");
  NXT_printNumber(aggregatedInbetweenTime); NXT_printString("\n");
}

void GarbageCollection_initialize() {
  endOfRAMObjects = NXTMemory_getStart();
  GarbageCollection_resetStatistics();
}

static inline ObjectRef Object_getReference(ObjectRef object, int number) {
  if (number == 0) {
    return Object_getClass(object);
  } else {
    return Object_getInstVar(object, number - 1);
  }
}

static inline void Object_setReference(ObjectRef object, int number, ObjectRef value) {
  if (number == 0) {
    Object_getForWrite(object)->klass = value;
  } else {
    Object* objectPtr = Object_get(object);
    ObjectRef* instVar = Object_isIndexable(object) ? &((IndexableObject*)objectPtr)->instanceVariables[number - 1] : &objectPtr->instanceVariables[number - 1];
    *instVar = value;
  }
}

static inline U16 Object_getReferenceCount(ObjectRef object) {
  if (Object_getElementSize(object) != 2) {
    if (Method_isMethod(Method_get(object))) {
      if (Method_isPrimitive(Method_get(object))) {
        return 0;
      }
      return Method_getLiteralCount(Method_get(object)) + 2;
    }
    return 1;
  }
  return Object_getSize(object) + 1;
}

#define NoObject 0

static void GarbageCollection_markRAMObject(ObjectRef rootObject);

static void GarbageCollection_markProcess(ObjectRef object) {
  ObjectTable_mark(object);
  Process* process = Process_get(object);
  GarbageCollection_markRAMObject(process->nextLink);
  GarbageCollection_markRAMObject(process->currentFrame);
  GarbageCollection_markRAMObject(process->waitsFor);
  ObjectTable_mark(process->stack);
  U16 stackTop = Process_getStackTopIndex(process);
  for (int i = 0; i <= stackTop; i++) {
    GarbageCollection_markRAMObject(Object_getReference(process->stack, i));
  }
}

#if 0
static void GarbageCollection_markRAMObject(ObjectRef rootObject) {
  // only start, if we have an RAM object, which is not already marked
  if (ObjectMemory_isRAMObject(rootObject) && !Object_isMarked(rootObject)) {
    /*if (ObjectMemory_isRAMObject(rootObject)) {
      Object_incReferenceCount(rootObject);
    }*/
    ObjectRef next = rootObject;
    ObjectRef back = NoObject;
    Object_mark(next);
    int i = 0;
    while (TRUE) {
      // do we have an object with object references and is our
      // field index within?
      if (i < Object_getReferenceCount(next)) {
        ObjectRef field = Object_getReference(next,i);
        /*if (ObjectMemory_isRAMObject(field)) {
          Object_incReferenceCount(field);
        }*/
        if (ObjectMemory_isRAMObject(field) && !Object_isMarked(field)) {
          // step down in the object graph
          Object_setReference(next, i, back);
          back = next;
          next = field;
          Object_mark(next);
          Interpreter_stackPush(i);
          i = 0;
        } else {
          // next instance field
          i++;
        }
      } else {
        // we're finished with this object, go back
        ObjectRef field = next;
        next = back;

        if (next == (ObjectRef)NoObject) {
          break;
        }

        // go up and correct instance field
        i = Interpreter_stackPop();
        back = Object_getReference(next, i);
        Object_setReference(next, i, field);
        // next instance field
        i++;
      }
    }
  }
}
#else
static void GarbageCollection_markRAMObject(ObjectRef rootObject) {
  // only start, if we have an RAM object, which is not already marked
  if (ObjectTable_isRAMObject(rootObject) && !ObjectTable_isMarked(rootObject)) {
    if (Object_getClass(rootObject) == specialObjects->NXTProcess) {
      GarbageCollection_markProcess(rootObject);
      return;
    }
    if (Object_getClass(rootObject) == specialObjects->NXTWeakArray) {
      // don't mark elements
      ObjectTable_mark(rootObject);
      return;
    }
    ObjectTable_mark(rootObject);
    // mark all subfields
    U16 referenceCount = Object_getReferenceCount(rootObject);
    for (int i = 0; i < referenceCount; i++) {
      GarbageCollection_markRAMObject(Object_getReference(rootObject,i));
    }
  }
}
#endif

static void GarbageCollection_markObject(ObjectRef rootObject) {
  if (!ObjectTable_isMarked(rootObject)) {
    if (Object_getClass(rootObject) == specialObjects->NXTWeakArray) {
      // don't mark elements
      ObjectTable_mark(rootObject);
      return;
    }
    ObjectTable_mark(rootObject);
    markedObjects++;
    // mark all subfields
    U16 referenceCount = Object_getReferenceCount(rootObject);
    for (int i = 0; i < referenceCount; i++) {
      GarbageCollection_markObject(Object_getReference(rootObject,i));
    }
  }
}


static U16 GarbageCollection_calculateFullSizeForEncodedObject(ObjectRef reference, Object* address) {
  if (ObjectTable_isIndexable(reference)) {
    U16 size = ((IndexableObject*)address)->size * ObjectTable_decodeElementSize(reference) + 2 + 2;
    size += size & 0x1;
    return size;
  } else {
    return ObjectTable_decodeSizeForNonIndexableObject(reference) * 2 + 2;
  }
}

static U16 GarbageCollection_calculateFreeEntrySize(ObjectRef reference, FreeEntry* address) {
  if (Freelist_isVerySmall(reference)) {
    return 2;
  } else {
    return address->size;
  }
}

static void GarbageCollection_compactRAM() {
  Object* source = (Object*)NXTMemory_getStart();
  Object* target = source;
  int ramObjects = 0;

  // exchange for every ram object class reference with object table index
  for (int i = minReference; i <= MAX_REFERENCE; i += 2) {
    ObjectRef reference = i;
    if (ObjectTable_get(reference) != 0 && ObjectTable_isRAMObject(reference)) {
      Object* object = ObjectTable_getAddress(reference);
      ramObjects++;
      if (ObjectTable_isFree(reference)) {
        // free list entry, object table gets NULL to indicate its not used
        FreeEntry* free = (FreeEntry*)object;
        //objectTable[reference >> 1] = 0;
        free->next = reference;
      } else {
        // real object, save its class in object table and put table index into object
        if (!Object_isClass(reference)) {
          ObjectTable_encodeClassAndSizeInfo(reference);
          object->klass = reference;
        }
      }
    }
  }

  int removedObjects = 0;
  while ((U8*)source < endOfRAMObjects) {
    U16 sourceRef = source->klass;
    if (!ObjectTable_isEncoded(sourceRef)) {
      // find sourceRef by scanning
      for (int i = minReference; i <= MAX_REFERENCE; i += 2) {
        if (ObjectTable_get(i) != 0 && !ObjectTable_isEncoded(i) && ObjectTable_getAddress(i) == source) {
          sourceRef = i;
          break;
        }
      }
    }
    if (ObjectTable_isFree(sourceRef)) {
      // free block
      source = (Object*)((U8*)source + GarbageCollection_calculateFreeEntrySize(sourceRef, (FreeEntry*)source));
      objectTable[sourceRef >> 1] = 0;
    } else {
      U16 sourceSize;
      if (ObjectTable_isEncoded(sourceRef)) {
        sourceSize = GarbageCollection_calculateFullSizeForEncodedObject(sourceRef, source);
      } else {
        sourceSize = Object_getFullSize(sourceRef);
      }
      // store following source pointer, because the source could be overwritten
      Object* nextSource = (Object*)((U8*)source + sourceSize);
      if (ObjectTable_isMarked(sourceRef) || ObjectTable_isRenewed(sourceRef)) {
        // unmark every marked object
        ObjectTable_setMarked(sourceRef, FALSE);
        if (ObjectTable_isEncoded(sourceRef)) {
          // restore class
          source->klass = ObjectTable_decodeClassInfo(sourceRef);
        }
        if (source != target) {
          // move object and adjust object table entry
          NXTMemory_move(target, source, sourceSize);
        }
        ObjectTable_setAddress(sourceRef, target);
        ObjectTable_setFree(sourceRef, FALSE);
        // adjust target
        target = (Object*)((U8*)target + sourceSize);
      } else {
        // free object table entry
        objectTable[sourceRef >> 1] = 0;
        removedObjects++;
      }
      source = nextSource;
    }
  }
  endOfRAMObjects = (U8*)target;
  objectCount -= removedObjects;
  Freelist_initialize();
}

/** \brief Compacts everything to Flash. */
static void GarbageCollection_compact() {
  U8 targetBuf[NXTFlash_getPageSize()];
  Object* target = Image_getFirstROMObject();
  U16 targetBufPos = (U32)target % NXTFlash_getPageSize();
  U32 targetPage = NXTFlash_firstPage();
  struct {
    U16 reference;
    U8* address;
  } addressBuffer[128];
  U8  addressCount = 0;
  bool lastPageWrittenAndOverlap = FALSE;

  // Store cursor position for displaying process
  U8 cursorX = NXTDisplay_getCursorX();
  U8 cursorY = NXTDisplay_getCursorY();

  // fill buf with image header and special objects
  NXTMemory_copy(targetBuf, NXTFlash_getStart(), targetBufPos);

  objectCount = 0;
  // First loop through Flash objects
  while (TRUE) {
    U16 sourceRef = 0;
    Object* source = (void*)-1;
    // search next nearest live ROM object
    for (int ref = minReference; ref <= MAX_REFERENCE; ref += 2) {
      if (ObjectTable_isMarked(ref) && ObjectTable_isROMObject(ref)) {
        Object* object = Object_get(ref);
        if (object >= target) {
          if (object < source) {
            sourceRef = ref;
            source = object;
          }
        }
      }
    }
    if (sourceRef == 0) {
      // No more ROM objects to process left
      break;
    }
    if (source != target) {
      addressBuffer[addressCount].reference = sourceRef;
      addressBuffer[addressCount].address = (U8*)target;
      addressCount++;
    }
    U32 size = Object_getFullSize(sourceRef);
    Object* newTarget = (Object*)((U8*)target + size);
    // move object to buffer and/or flash
    U32 toCopy = size;
    U32 copied = 0;
    while (toCopy > 0) {
      U32 toCopyThisTime = MIN(toCopy, NXTFlash_getPageSize() - targetBufPos);
      NXTMemory_copy(targetBuf + targetBufPos, (U8*)source + copied, toCopyThisTime);
      toCopy -= toCopyThisTime;
      targetBufPos += toCopyThisTime;
      copied += toCopyThisTime;
      if (targetBufPos == NXTFlash_getPageSize()) {
        // Write buffer to current Flash page only if we move some objects here
        if (addressCount > 0 || lastPageWrittenAndOverlap) {
          NXTFlash_writePage(targetBuf, targetPage);
          if (toCopy > 0) {
            lastPageWrittenAndOverlap = TRUE;
          } else {
            lastPageWrittenAndOverlap = FALSE;
          }
          // Use address buffer to update object addresses
          for (int j = 0; j < addressCount; j++) {
            ObjectTable_setAddress(addressBuffer[j].reference, addressBuffer[j].address);
          }
        }
        targetPage++;
        targetBufPos = 0;
        addressCount = 0;
      }
    }
    target = newTarget;
    objectCount++;
    NXTDisplay_setCursor(cursorX, cursorY);
    NXT_printNumber(objectCount * 100 / markedObjects);
    NXT_printString("%");
  }

  // now loop through RAM objects
  for (int sourceRef = minReference; sourceRef <= MAX_REFERENCE; sourceRef += 2) {
    if (ObjectTable_isMarked(sourceRef) && ObjectTable_isRAMObject(sourceRef)) {
      Object* source = Object_get(sourceRef);
      U32 size = Object_getFullSize(sourceRef);
      Object* newTarget = (Object*)((U8*)target + size);
      // move object to buffer and/or flash
      U32 toCopy = size;
      U32 copied = 0;
      while (toCopy > 0) {
        U32 toCopyThisTime = MIN(toCopy, NXTFlash_getPageSize() - targetBufPos);
        NXTMemory_copy(targetBuf + targetBufPos, (U8*)source + copied, toCopyThisTime);
        toCopy -= toCopyThisTime;
        targetBufPos += toCopyThisTime;
        copied += toCopyThisTime;
        if (targetBufPos == NXTFlash_getPageSize()) {
          // Write buffer to current Flash page only if we move some objects here
          NXTFlash_writePage(targetBuf, targetPage);
          // Use address buffer to update object addresses
          for (int j = 0; j < addressCount; j++) {
            ObjectTable_setAddress(addressBuffer[j].reference, addressBuffer[j].address);
          }
          targetPage++;
          targetBufPos = 0;
          addressCount = 0;
        }
      }
      addressBuffer[addressCount].reference = sourceRef;
      addressBuffer[addressCount].address = (U8*)target;
      addressCount++;
      target = newTarget;
      objectCount++;
      NXTDisplay_setCursor(cursorX, cursorY);
      NXT_printNumber(objectCount * 100 / markedObjects);
      NXT_printString("%");
    }
  }

  // finally, flush buffer
  if (targetBufPos > 0) {
    NXTFlash_writePage(targetBuf, targetPage);
    for (int i = 0; i < addressCount; i++) {
      ObjectTable_setAddress(addressBuffer[i].reference, addressBuffer[i].address);
    }
  }

  // clean up object table
  for (int ref = minReference; ref <= MAX_REFERENCE; ref += 2) {
    if (!ObjectTable_isMarked(ref)) {
      ObjectTable_set(ref, 0);
    } else {
      NXT_ASSERT(ObjectTable_isRealObject(Object_get(ref)->klass), "Blub");
      NXT_ASSERT(!ObjectTable_isRealObject(Class_get(Object_get(ref)->klass)->format), "Blub");
      NXT_ASSERT(Object_get(Class_get(Object_get(ref)->klass)->name)->klass == specialObjects->NXTSymbol, "Blub");
    }
  }

  NXTDisplay_setCursor(cursorX, cursorY);
  NXT_printString("Done.\n");
}

void GarbageCollection_collect() {
  GarbageCollection_updateStatistics(TRUE);
  // Advise the interpreter to store its execution state properly in
  // the Smalltalk objects
  Interpreter_storeExecutionState();

  // mark current stack so its not traversed (we use it ourselves! :) )
  ObjectTable_mark(Interpreter_stack());
  // mark elements of current stack
  int stackTop = Interpreter_stackTopIndex();
  for (int i = 0; i <= stackTop; i++) {
    GarbageCollection_markRAMObject(Interpreter_stackAt(i));
  }
  // mark primitive temporary set
  for (int i = 0; i < temporarySetSize; i++) {
    GarbageCollection_markRAMObject(temporarySet[i]);
  }
  // mark current frame (TODO: eleminate this)
  GarbageCollection_markRAMObject(currentFrame);
  // mark frame pool
  for (int i = 0; i <= MAX_FRAME_SIZE; i++) {
    GarbageCollection_markRAMObject(Frame_getPool()[i]);
  }
  // mark special objects
  ObjectRef* objPtr = (ObjectRef*)specialObjects;
  while (objPtr < (ObjectRef*)((U8*)specialObjects + sizeof(SpecialObjects) / sizeof (ObjectRef))) {
    GarbageCollection_markRAMObject(*objPtr);
    objPtr++;
  }
  // TODO: Eliminate this mark phase by using a rootset instead of the renewed bit (or find something better)
  for (int i = minReference; i <= MAX_REFERENCE; i += 2) {
    if (ObjectTable_isRenewed(i)) {
      GarbageCollection_markRAMObject(i);
    }
  }
  // Compact phase
  GarbageCollection_compactRAM();

  // Shrink object table as far as possible
  while (minReference < MAX_REFERENCE && objectTable[minReference >> 1] == 0) {
    minReference += 2;
    objectTableStart++;
    objectTableROMStart++;
  }

  // Nil out weak pointers without target
  // TODO: This can be done during compacting phase
  ObjectRef weakArray = Class_findFirstInstance(specialObjects->NXTWeakArray);
  while (weakArray != specialObjects->nil) {
    IndexableObject* weakArrayPtr = IndexableObject_get(weakArray);
    int size = weakArrayPtr->size;
    for (int i = 0; i < size; i++) {
      ObjectRef element = weakArrayPtr->instanceVariables[i];
      if (ObjectTable_get(element) == 0) {
        weakArrayPtr->instanceVariables[i] = specialObjects->nil;
      }
    }
    weakArray = ObjectMemory_findNextInstanceFor(weakArray);
  }

  Object_incReferenceCount(currentFrame);
  // Inform the interpreter, that objects have moved
  Interpreter_reloadExecutionState();

  GarbageCollection_updateStatistics(FALSE);
}

void GarbageCollection_finalCollect() {
  // here, we only need to mark special objects, as every thing needed is linked by them
  markedObjects = 0;
  ObjectRef* objPtr = (ObjectRef*)specialObjects;
  while (objPtr < (ObjectRef*)((U8*)specialObjects + sizeof(SpecialObjects) / sizeof (ObjectRef))) {
    GarbageCollection_markObject(*objPtr);
    objPtr++;
  }

  // Compact everything to the start of the Flash
  GarbageCollection_compact();

  // Shrink object table as far as possible
  while (minReference < MAX_REFERENCE && objectTable[minReference >> 1] == 0) {
    minReference += 2;
    objectTableStart++;
    objectTableROMStart++;
  }
  // TODO: Nil out weak pointers during compaction
}
