/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

/** \file garbagecollection.h
 * \author Martin Beck
 * \brief Functions and variables for garbage collection.
 *
 * The garbage collection in NXTalk is mainly a Mark & Compact algorithm,
 * which takes place if the reference counting and free list management
 * isn't enough anymore. Thus, free space is almost completely used up.
 * The mark phase simply marks all objects reachable from a root set
 * recursively. Afterwards the marked objects are moved to the beginning
 * of the RAM leaving dead ones alone. Thus, the object heap is compacted
 * and more free RAM might be available again.
 */

#ifndef _GARBAGECOLLECTION_H_
#define _GARBAGECOLLECTION_H_

#include "object.h"

/** \brief Initialize the garbage collection module. */
void GarbageCollection_initialize(void);

/** \brief Prints garbage collection statistics on the display. */
void GarbageCollection_printStatistics(void);

/** \brief Resets the garbage collection statistics. */
void GarbageCollection_resetStatistics(void);

/** \brief Starts a garbage collection. */
void GarbageCollection_collect(void);

/** \brief Full garbage collection for shutting down the NXT.
 *
 * This function does a full garbage collection over RAM and ROM. It does
 * not include the current stack etc. as it assumes, the NXT will be shutdown
 * and only special objects like classes or packages are needed. Thus, do
 * not call this function if the system is still running.
 */
void GarbageCollection_finalCollect(void);

#endif // _GARBAGECOLLECTION_H_
