/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

#include "image.h"

#include "garbagecollection.h"
#include "freelist.h"
#include "class.h"

SpecialObjects* specialObjects;
U8* endOfRAMObjects;
U8* endOfROMObjects;

bool Image_isPresent() {
  return *(U16*)NXTFlash_getStart() == 0x4242;
}

Object* Image_getFirstROMObject() {
  return (Object*)(NXTFlash_getStart() + 2 * sizeof(U16) + sizeof(SpecialObjects));
}

void Image_initializeObjectMemory() {
  Freelist_initialize();
  U16* pointer = (U16*)NXTFlash_getStart();
  U16 header = *pointer++;
  if (header == 0x4242) {
    objectCount = *pointer++;
    specialObjects = (SpecialObjects*)pointer;

    // Copy and relocate object table
    U32* objectTableEntry = (U32*)(NXTFlash_getEnd() - 4);
    minReference = 65535;
    objectTable = (U32*)(NXTMemory_getEnd() - 4 - (65535 >> 1) * 4);
    objectTableStart = (U32*)(NXTMemory_getEnd() - objectCount * 4);
    objectTableROMStart = (U32*)NXTFlash_getEnd() - objectCount * 4;
    for (int i = 0; i < objectCount; i++) {
      U32 offset = *objectTableEntry;
      if (offset != 0) {
        Object* object = (Object*)(NXTFlash_getStart() + offset);
        ObjectTable_setAddress(minReference, object);
        ObjectTable_setFree(minReference, FALSE);
        ObjectTable_setReferenceCount(minReference, RC_TOO_MANY_REFERENCES);
        ObjectTable_setMarked(minReference, FALSE);
      } else {
        ObjectTable_set(minReference, 0);
      }
      objectTableEntry--;
      minReference -= 2;
    }
    minReference += 2;
    for (int reference = minReference; reference < MAX_REFERENCE; reference += 2) {
      void* address = ObjectTable_getAddress(reference);
      if (address != NULL) {
        endOfROMObjects = MAX(endOfROMObjects, (U8*)address);
        ObjectTable_setIndexable(reference, Class_isVariable(Class_get(Object_getClass(reference))));
        ObjectTable_setReferencingInstances(reference, Class_getElementSize(Class_get(Object_getClass(reference))) == 2);
      }
    }
    GarbageCollection_initialize();
  } else {
    NXT_error("No image found!");
  }
}

void Image_updateHeader(void) {
  U8 page[NXTFlash_getPageSize()];
  NXTFlash_readPage(NXTFlash_firstPage(), page);
  ((U16*)page)[1] = (MAX_REFERENCE - minReference) / 2 + 1;
  NXTFlash_writePage(page, NXTFlash_firstPage());
}

void Image_storeObjectTable(void) {
  U8 entriesPerPage = NXTFlash_getPageSize() / 4;
  U32 page[entriesPerPage];
  U32 pageNumber = NXTFlash_lastPage();
  U16 reference = MAX_REFERENCE;
  while (reference - entriesPerPage * 2 > minReference) {
    for (int i = entriesPerPage - 1; i >= 0; i--) {
      if (ObjectTable_get(reference) != 0) {
        U8* address = (U8*)ObjectTable_getAddress(reference);
        page[i] = address - NXTFlash_getStart();
      } else {
        page[i] = 0;
      }
      reference -= 2;
    }
    NXTFlash_writePage((U8*)page, pageNumber--);
  }

  NXTFlash_readPage(pageNumber, (U8*)page);
  int i = entriesPerPage - 1;
  while (reference >= minReference) {
    if (ObjectTable_get(reference) != 0) {
      U8* address = (U8*)ObjectTable_getAddress(reference);
      page[i] = address - NXTFlash_getStart();
    } else {
      page[i] = 0;
    }
    i--;
    reference -= 2;
  }
  NXTFlash_writePage((U8*)page, pageNumber);
}

void* Image_allocateRAM(U32 size) {
  // Size should be even for NXT
  size += size & 0x1;
  if (endOfRAMObjects + size  > (U8*)objectTableStart) {
    return NULL;
  }
  U8* newObject = endOfRAMObjects;
  // Move end-of-objects pointer to the next even address
  endOfRAMObjects += size;
  return newObject;
}

void Image_moveObjectToRAM(ObjectRef object) {
  U16 objectSize = Object_getFullSize(object);
  Object* newObject = Image_allocateRAM(objectSize);
  if (newObject == NULL) {
    GarbageCollection_collect();
    newObject = Image_allocateRAM(objectSize);
    if (newObject == NULL) {
      NXT_error("Out of RAM");
    }
  }
  NXTMemory_copy(newObject, ObjectTable_getAddress(object), objectSize);
  ObjectTable_setAddress(object, newObject);
  ObjectTable_setRenewed(object, TRUE);
}
