/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

/** \file image.h
 * \author Martin Beck
 * \brief Functions and variables for the Smalltalk image.
 */

#ifndef _IMAGE_H_
#define _IMAGE_H_

#include "object.h"
#include "integer.h"
#include "objecttable.h"

/** \brief Describes a list of special objects.
 *
 * Some objects have to be known to the virtual machine at any time and
 * are thus special. This includes a reference to the nil, true or false
 * singletons as well as refrences to some classes from which instances
 * need to be created.
 */
struct SpecialObjects {
  /** \brief The symbol to send to the system object to start the system. */
  ObjectRef startupSymbol;

  /** \brief A reference to the system object. */
  ObjectRef system;

  /** \brief The nil object. */
  ObjectRef nil;

  /** \brief The true object. */
  ObjectRef true;

  /** \brief The false object. */
  ObjectRef false;

  /** \brief A reference to the sole NXTProcessorScheduler instance. */
  ObjectRef Processor;

  /** \brief The system dictionary object. */
  ObjectRef Smalltalk;

  /** \brief The symbol #doesNotUnderstand: for DNU message creation.*/
  ObjectRef doesNotUnderstand;

  /** \brief The symbol #onNonLocalReturnsDo: for catching NLRs. */
  ObjectRef onNonLocalReturnsDo;

  /** \brief A reference to the NXTSmallInteger class. */
  ObjectRef NXTSmallInteger;

  /** \brief A reference to the NXTLargePositiveInteger class. */
  ObjectRef NXTLargePositiveInteger;

  /** \brief A reference to the NXTLargeNegativeInteger class. */
  ObjectRef NXTLargeNegativeInteger;

  /** \brief A reference to the NXTByteArray class. */
  ObjectRef NXTByteArray;

  /** \brief A reference to the NXTMethod class. */
  ObjectRef NXTMethod;

  /** \brief A reference to the NXTBlock class. */
  ObjectRef NXTBlock;

  /** \brief A reference to the NXTSymbol class. */
  ObjectRef NXTSymbol;

  /** \brief A reference to the NXTMethodFrame class. */
  ObjectRef NXTMethodFrame;

  /** \brief A reference to the NXTBlockFrame class. */
  ObjectRef NXTBlockFrame;

  /** \brief A reference to the NXTArray class. */
  ObjectRef NXTArray;

  /** \brief A reference to the NXTProcess class. */
  ObjectRef NXTProcess;

  /** \brief A reference to the NXTMessage class for DNU message creation. */
  ObjectRef NXTMessage;

  /** \brief The NXTWeakArray class for recognizing weak references. */
  ObjectRef NXTWeakArray;

  /** \brief A reference to the NXTClass class. */
  ObjectRef NXTClass;

  /** \brief A reference to the NXTMetaclass class. */
  ObjectRef NXTMetaclass;
} __attribute__ ((packed));
typedef struct SpecialObjects SpecialObjects;

/** \brief The special objects in the current image. */
extern SpecialObjects* specialObjects;

/** \brief The end of the object heap. */
extern U8* endOfRAMObjects;

/** \brief The end of the ROM object heap. */
extern U8* endOfROMObjects;

/** \brief Checks, whether an NXTalk image is present.
 *
 * \return TRUE, if an image is present. FALSE, if not.
 */
bool Image_isPresent(void);

/** \brief Initializes the image and its object memory. */
void Image_initializeObjectMemory(void);

/** \brief Returns the address of the first object in ROM. */
Object* Image_getFirstROMObject(void);

/** \brief Updates the object count in the Flash header. */
void Image_updateHeader(void);

/** \brief Stores the current object table and its meta-data in Flash. */
void Image_storeObjectTable(void);

/** \brief Allocates memory in the object heap in the RAM.
 *
 * This memory is always allocated at the end of the object heap, even
 * if some objects were already freed by the reference counting. Hence,
 * the object heap will always grow with this operation.
 *
 * \param size in bytes of the needed RAM.
 * \return A pointer to the allocated space. NULL, if the heap in RAM
 *         cannot grow anymore.
 */
void* Image_allocateRAM(U32 size);

/** \brief Moves an object from the Flash ROM to the RAM.
 *
 * This function ensures, that the given object lies in RAM after
 * execution, doing a garbage collection if necessary. Fails with
 * a system halt, if the object is still too large for the free
 * RAM.
 *
 * \param object The object to move to RAM.
 */
void Image_moveObjectToRAM(ObjectRef object);

#endif //_IMAGE_H_
