/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

#include "integer.h"

#include <stdlib.h>

#include "garbagecollection.h"
#include "primitives.h"
#include "image.h"
#include "boolean.h"
#include "class.h"

static bool LargeInteger_isLargeInteger(ObjectRef object) {
  if (Object_getClass(object) == specialObjects->NXTLargePositiveInteger ||
      Object_getClass(object) == specialObjects->NXTLargeNegativeInteger) {
    return TRUE;
      } else {
        return FALSE;
      }
}

static S32 LargeInteger_fromReference(LargeInteger value) {
  U16 size = Object_getSize(value);
  S32 result = 0;
  if (size > 4) {
    NXT_error("Too large int");
  }
  for (int i = 0; i < size; i++) {
    result += ByteObject_get(value)->bytes[i] << (i * 8);
  }
  if (Object_getClass(value) == specialObjects->NXTLargeNegativeInteger) {
    return -result;
  } else {
    return result;
  }
}

S32 Integer_fromReference(ObjectRef number) {
  if (SmallInteger_isSmallInteger(number)) {
    return SmallInteger_fromReference(number);
  } else if (LargeInteger_isLargeInteger(number)) {
    return LargeInteger_fromReference(number);
  }
  NXT_error("No integer");
  return 0;
}

LargeInteger LargeInteger_new(S32 value) {
  ClassRef class;
  if (value >= 0) {
    class = specialObjects->NXTLargePositiveInteger;
  } else {
    value = -value;
    class = specialObjects->NXTLargeNegativeInteger;
  }
  U8 bytes[4];
  U8 size = 0;
  while (value > 0) {
    bytes[size++] = value & 0xFF;
    value >>= 8;
  }
  LargeInteger result = Primitive_createIntermediateInstance(class, size);
  ByteObject* resultPtr = ByteObject_get(result);
  for (int i = 0; i < size; i++) {
    resultPtr->bytes[i] = bytes[i];
  }
  return result;
}

static bool LargeInteger_isPositive(LargeInteger operand) {
  return specialObjects->NXTLargePositiveInteger == Object_getClass(operand);
}

static LargeInteger LargeInteger_normalize(LargeInteger operand) {
  U16 opSize = Object_getSize(operand);
  ByteObject* opPtr = ByteObject_get(operand);
  U16 lastUsedDigit = opSize - 1;
  while (lastUsedDigit > 1 && opPtr->bytes[lastUsedDigit] == 0) {
    lastUsedDigit--;
  }
  if (lastUsedDigit <= 1) {
    // maybe small integer, otherwise cannot get smaller
    U8 digit1 = opPtr->bytes[1];
    U8 digit0 = opPtr->bytes[0];
    if (LargeInteger_isPositive(operand)) {
      // positive max value for SmallInteger is 16r3FFF
      if (digit1 < 0x40) {
        return SmallInteger_toReference((digit1 << 8) | digit0);
      }
    } else {
      // negative, min value for SmallInteger is -16r4000
      if (digit1 < 0x40 || (digit1 == 0x40 && digit0 == 0x00)) {
        return SmallInteger_toReference((digit1 << 8) | digit0);
      }
    }
  }
  if (opSize - (lastUsedDigit + 1) > 0) {
    LargeInteger result = Primitive_createIntermediateInstance(Object_getClass(operand), lastUsedDigit + 1);
    for (int i = 0; i <= lastUsedDigit; i++) {
      Object_setInstVar(result, i, Object_getInstVar(operand, i));
    }
    return result;
  }
  // Nothing to normalize
  return operand;
}

static LargeInteger LargeInteger_add(LargeInteger operand1, LargeInteger operand2) {
  U16 op1Size = Object_getSize(operand1);
  U16 op2Size = Object_getSize(operand2);
  ByteObject* op1Ptr = ByteObject_get(operand1);
  ByteObject* op2Ptr = ByteObject_get(operand2);
  U16 newSize = MAX(op1Size, op2Size) + 1;
  LargeInteger result = Primitive_createIntermediateInstance(specialObjects->NXTLargePositiveInteger, newSize);
  ByteObject* resultPtr = ByteObject_get(result);
  U16 carry = 0;
  for (int i = 0; i < newSize; i++) {
    U16 op1 = (i < op1Size) ? op1Ptr->bytes[i] : 0;
    U16 op2 = (i < op2Size) ? op2Ptr->bytes[i] : 0;
    U16 res = op1 + op2 + carry;
    resultPtr->bytes[i] = res & 0xFF;
    carry = res >> 8;
  }
  return LargeInteger_normalize(result);
}

static LargeInteger LargeInteger_sub(LargeInteger operand1, LargeInteger operand2) {
  U16 op1Size = Object_getSize(operand1);
  U16 op2Size = Object_getSize(operand2);
  ByteObject* op1Ptr = ByteObject_get(operand1);
  ByteObject* op2Ptr = ByteObject_get(operand2);
  U16 newSize = MAX(op1Size, op2Size);
  LargeInteger result = Primitive_createIntermediateInstance(specialObjects->NXTLargePositiveInteger, newSize);
  ByteObject* resultPtr = ByteObject_get(result);
  U16 carry = 0;
  for (int i = 0; i < newSize; i++) {
    U16 op1 = (i < op1Size) ? op1Ptr->bytes[i] : 0;
    U16 op2 = (i < op2Size) ? op2Ptr->bytes[i] : 0;
    S16 res = op1 - (op2 + carry);
    carry = 0;
    while (res < 0) {
      res += 256;
      carry++;
    }
    resultPtr->bytes[i] = res & 0xFF;
  }
  return LargeInteger_normalize(result);
}

static LargeInteger LargeInteger_mul(LargeInteger operand1, LargeInteger operand2) {
  U16 op1Size = Object_getSize(operand1);
  U16 op2Size = Object_getSize(operand2);
  U16 newSize = op1Size + op2Size;
  ByteObject* op1Ptr = ByteObject_get(operand1);
  ByteObject* op2Ptr = ByteObject_get(operand2);
  LargeInteger prod = Primitive_createIntermediateInstance(specialObjects->NXTLargePositiveInteger, newSize);
  ByteObject* prodPtr = ByteObject_get(prod);
  for (int i = 0; i < op1Size; i++) {
    U16 digit = op1Ptr->bytes[i];
    if (digit != 0) {
      U16 k = i;
      int carry = 0;
      for (int j = 0; j < op2Size; j++) {
        int ab = op2Ptr->bytes[j] * digit + carry + prodPtr->bytes[k];
        carry = ab >> 8;
        prodPtr->bytes[k] = ab & 0xFF;
        k++;
      }
      prodPtr->bytes[k] = carry;
    }
  }
  return LargeInteger_normalize(prod);
}

static bool LargeInteger_lessThan(ObjectRef operand1, ObjectRef operand2) {
  // both have to be large integer
  U16 op1Size = Object_getSize(operand1);
  U16 op2Size = Object_getSize(operand2);
  U16 maxSize = MAX(op1Size, op2Size);
  ByteObject* op1Ptr = ByteObject_get(operand1);
  ByteObject* op2Ptr = ByteObject_get(operand2);
  for (int i = maxSize - 1; i >= 0; i--) {
    U16 op1 = (i < op1Size) ? op1Ptr->bytes[i] : 0;
    U16 op2 = (i < op2Size) ? op2Ptr->bytes[i] : 0;
    if (op1 < op2) {
      return TRUE;
    }
    if (op1 > op2) {
      return FALSE;
    }
  }
  return FALSE;
}

static bool LargeInteger_equal(ObjectRef operand1, ObjectRef operand2) {
  // both have to be large integer
  U16 op1Size = Object_getSize(operand1);
  U16 op2Size = Object_getSize(operand2);
  U16 maxSize = MAX(op1Size, op2Size);
  ByteObject* op1Ptr = ByteObject_get(operand1);
  ByteObject* op2Ptr = ByteObject_get(operand2);
  for (int i = maxSize - 1; i >= 0; i--) {
    U16 op1 = (i < op1Size) ? op1Ptr->bytes[i] : 0;
    U16 op2 = (i < op2Size) ? op2Ptr->bytes[i] : 0;
    if (op1 < op2) {
      return FALSE;
    }
    if (op1 > op2) {
      return FALSE;
    }
  }
  return TRUE;
}

/********** primitives ********/

void SmallInteger_primPrint(PrimitiveFrame* frame) {
  ObjectRef argument = Primitive_argumentAt(frame, 0);
  SmallInteger number = SmallInteger_fromReference(argument);
  if (number < 0) {
    NXT_printString("-");
    number = -number;
  }
  NXT_printNumber(number);
  Primitive_return(frame, argument);
}

void SmallInteger_primAdd(PrimitiveFrame* frame) {
  SmallInteger smallInt1 = frame->self;
  SmallInteger smallInt2 = Primitive_argumentAt(frame, 0);
  if ((int)SmallInteger_fromReference(smallInt1) + (int)SmallInteger_fromReference(smallInt2) > SMALLINTEGER_MAX) {
    Primitive_return(frame, Integer_new((int)SmallInteger_fromReference(smallInt1) + (int)SmallInteger_fromReference(smallInt2)));
    return;
  }
  if (!(SmallInteger_isSmallInteger(smallInt2))) {
    NXT_error("No SmallInts!");
  }
  Primitive_return(frame, smallInt1 + smallInt2);
}

void SmallInteger_primSub(PrimitiveFrame* frame) {
  SmallInteger smallInt1 = frame->self;
  SmallInteger smallInt2 = Primitive_argumentAt(frame, 0);
  Primitive_return(frame, smallInt1 - smallInt2);
}

void SmallInteger_primMul(PrimitiveFrame* frame) {
  SmallInteger argument = Primitive_argumentAt(frame, 0);
  if (!SmallInteger_isSmallInteger(argument)) {
    if (LargeInteger_isLargeInteger(argument)) {
      Primitive_return(frame, LargeInteger_mul(LargeInteger_new(SmallInteger_fromReference(frame->self)), argument));
      return;
    } else {
      NXT_error("No integer");
    }
  }
  if ((int)SmallInteger_fromReference(argument) * (int)SmallInteger_fromReference(frame->self) > SMALLINTEGER_MAX) {
    Primitive_return(frame, LargeInteger_new((int)SmallInteger_fromReference(argument) *
                                             (int)SmallInteger_fromReference(frame->self)));
    return;
  }
  Primitive_return(frame, ((int)frame->self * (int)argument) / 2);
}

void SmallInteger_primDivFloor(PrimitiveFrame* frame) {
  SmallInteger smallInt1 = frame->self;
  SmallInteger smallInt2 = Primitive_argumentAt(frame, 0);
  Primitive_return(frame, SmallInteger_toReference(smallInt1 / smallInt2));
}

void SmallInteger_primMod(PrimitiveFrame* frame) {
  SmallInteger argument = SmallInteger_fromReference(Primitive_argumentAt(frame, 0));
  ObjectRef value = SmallInteger_toReference(SmallInteger_fromReference(frame->self) % argument);
  Primitive_return(frame, value);
}

void SmallInteger_primLessThan(PrimitiveFrame* frame) {
  SmallInteger smallInt1 = frame->self;
  SmallInteger smallInt2 = Primitive_argumentAt(frame, 0);
  if (!(SmallInteger_isSmallInteger(smallInt2))) {
    NXT_error("No SmallInts!");
  }
  if (smallInt1 < smallInt2) {
    Primitive_return(frame, specialObjects->true);
  } else {
    Primitive_return(frame, specialObjects->false);
  }
}

void SmallInteger_primLessThanOrEqual(PrimitiveFrame* frame) {
  SmallInteger smallInt1 = frame->self;
  SmallInteger smallInt2 = Primitive_argumentAt(frame, 0);
  if (!(SmallInteger_isSmallInteger(smallInt2))) {
    NXT_error("No SmallInts!");
  }
  if (smallInt1 <= smallInt2) {
    Primitive_return(frame, specialObjects->true);
  } else {
    Primitive_return(frame, specialObjects->false);
  }
}

void SmallInteger_primBitAnd(PrimitiveFrame* frame) {
  ObjectRef smallInt1 = frame->self;
  ObjectRef smallInt2 = Primitive_argumentAt(frame, 0);
  if (!(SmallInteger_isSmallInteger(smallInt2))) {
    NXT_error("No SmallInts!");
  }
  Primitive_return(frame, smallInt1 & smallInt2);
}

void SmallInteger_primBitOr(PrimitiveFrame* frame) {
  ObjectRef smallInt1 = frame->self;
  ObjectRef smallInt2 = Primitive_argumentAt(frame, 0);
  if (!(SmallInteger_isSmallInteger(smallInt2))) {
    NXT_error("No SmallInts!");
  }
  Primitive_return(frame, smallInt1 | smallInt2);
}

void SmallInteger_primBitXor(PrimitiveFrame* frame) {
  ObjectRef smallInt1 = frame->self;
  ObjectRef smallInt2 = Primitive_argumentAt(frame, 0);
  if (!(SmallInteger_isSmallInteger(smallInt2))) {
    NXT_error("No SmallInts!");
  }
  Primitive_return(frame, smallInt1 ^ smallInt2);
}

void SmallInteger_primBitShift(PrimitiveFrame* frame) {
  S16 self = SmallInteger_fromReference(frame->self);
  ObjectRef argument = Primitive_argumentAt(frame, 0);
  if (!SmallInteger_isSmallInteger(argument)) {
    NXT_error("Argument is no small int!");
  }
  S16 shift = SmallInteger_fromReference(argument);

  if (shift >= 0) {
    Primitive_return(frame, SmallInteger_toReference(self << shift));
  } else {
    Primitive_return(frame, SmallInteger_toReference(self >> -shift));
  }
}

void SmallInteger_primAtRandom(PrimitiveFrame* frame) {
  S16 range = SmallInteger_fromReference(frame->self);
  if (range <= 0) {
    Primitive_return(frame, Integer_new(0));
  } else {
    Primitive_return(frame, Integer_new(NXT_rand() % range + 1));
  }
}

void LargeInteger_primAdd(PrimitiveFrame* frame) {
  LargeInteger operand = Primitive_argumentAt(frame, 0);
  if (SmallInteger_isSmallInteger(operand)) {
    Primitive_return(frame, LargeInteger_add(frame->self, LargeInteger_new(SmallInteger_fromReference(operand))));
    return;
  }
  if (Object_getClass(operand) != specialObjects->NXTLargePositiveInteger) {
    NXT_error("Wrong int format");
  }
  Primitive_return(frame, LargeInteger_add(frame->self, operand));
}

void LargeInteger_primSub(PrimitiveFrame* frame) {
  LargeInteger result = 0;
  LargeInteger operand = Primitive_argumentAt(frame, 0);
  if (SmallInteger_isSmallInteger(operand)) {
    LargeInteger largeOperand = LargeInteger_new(SmallInteger_fromReference(operand));
    result = LargeInteger_sub(frame->self, largeOperand);
  } else if (Object_getClass(operand) != specialObjects->NXTLargePositiveInteger) {
    NXT_error("Wrong int format");
  } else {
    result = LargeInteger_sub(frame->self, operand);
  }
  Primitive_return(frame, result);
}

void LargeInteger_primMul(PrimitiveFrame* frame) {
  LargeInteger operand = Primitive_argumentAt(frame, 0);
  if (SmallInteger_isSmallInteger(operand)) {
    Primitive_return(frame, LargeInteger_mul(frame->self, LargeInteger_new(SmallInteger_fromReference(operand))));
    return;
  }
  if (Object_getClass(operand) != specialObjects->NXTLargePositiveInteger) {
    NXT_error("Wrong int format");
  }
  Primitive_return(frame, LargeInteger_mul(frame->self, operand));
}

void LargeInteger_primLessThan(PrimitiveFrame* frame) {
  ObjectRef operand = Primitive_argumentAt(frame, 0);
  if (SmallInteger_isSmallInteger(operand)) {
    Primitive_return(frame, Boolean_new(LargeInteger_lessThan(frame->self, LargeInteger_new(SmallInteger_fromReference(operand)))));
  } else {
    Primitive_return(frame, Boolean_new(LargeInteger_lessThan(frame->self, operand)));
  }
}

void LargeInteger_primEquals(PrimitiveFrame* frame) {
  ObjectRef operand = Primitive_argumentAt(frame, 0);
  if (SmallInteger_isSmallInteger(operand)) {
    Primitive_return(frame, Boolean_new(LargeInteger_equal(frame->self, LargeInteger_new(SmallInteger_fromReference(operand)))));
  } else {
    Primitive_return(frame, Boolean_new(LargeInteger_equal(frame->self, operand)));
  }
}

void LargeInteger_primPrint(PrimitiveFrame* frame) {
  LargeInteger number = Primitive_argumentAt(frame, 0);
  if (Object_getClass(number) == specialObjects->NXTLargeNegativeInteger) {
    NXT_printString("-16r");
  } else {
    NXT_printString("16r");
  }
  U16 size = Object_getSize(number);
  for (int i = size - 1; i >= 0; i--) {
    U8 byte = ByteObject_get(number)->bytes[i];
    if (byte < 16) {
      NXT_printString("0");
    }
    NXT_printHexNumber(byte);
  }
  Primitive_return(frame, number);
}
