/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

/** \file integer.h
 * \author Martin Beck
 * \brief Functions and primitives for the integer classes.
 */

#ifndef _INTEGER_H_
#define _INTEGER_H_

#include "object.h"

/** \brief Type redefinition to indicate tagged SmallInteger "objects". */
typedef S16 SmallInteger;

/** \brief Type redefinition to indicate LargeInteger objects. */
typedef ObjectRef LargeInteger;

/** \brief Type redefinition to indicate Integer objects of any kind. */
typedef ObjectRef Integer;

/** \brief Minimum value for a tagged SmallInteger. */
#define SMALLINTEGER_MIN -16384

/** \brief Maximum value for a tagged SmallInteger. */
#define SMALLINTEGER_MAX 16383

/** \brief Indicates, whether the given object is a tagged SmallInteger.
 *
 * \param object The object to check.
 * \return TRUE, if it is a tagged integer. FALSE, if not.
 */
static inline bool SmallInteger_isSmallInteger(ObjectRef object) {
  return !(object & 0x1);
}

/** \brief Converts a native integer to a tagged one.
 *
 * \param number Native integer.
 * \return A tagged integer representing the value.
 */
static inline SmallInteger SmallInteger_toReference(S16 number) {
  return number << 1;
}

/** \brief Converts a tagged integer into a native one.
 *
 * \param number Reference to the tagged integer.
 * \return The native integer.
 */
static inline S16 SmallInteger_fromReference(SmallInteger number) {
  return (S16)number / 2;
}

/** \brief Returns the native value for any integer object.
 *
 * Does not work, if a LargeInteger is too large.
 *
 * \param number Reference to the LargeInteger or tagged SmallInteger.
 * \return Signed native value of the Integer.
 */
S32 Integer_fromReference(Integer number);

/** \brief Allocates a LargeInteger representing the given value.
 *
 * \param value The value for the LargeInteger.
 * \return A reference to the new LargeInteger object.
 */
LargeInteger LargeInteger_new(S32 value);

/** \brief Allocate an integer suitable for the value.
 *
 * Creates a tagged integer or a LargeInteger, depending on the given
 * value.
 *
 * \param value The native value to encode as object.
 * \return A tagged SmallInteger or a LargeInteger representing the given
 *         value.
 */
static inline Integer Integer_new(S32 value) {
  if (value < SMALLINTEGER_MIN || value > SMALLINTEGER_MAX) {
    return LargeInteger_new(value);
  } else {
    return SmallInteger_toReference(value);
  }
}

/** \name Primitives
 * @{
 */

/** \brief Primitive. Print the integer. */
void SmallInteger_primPrint(PrimitiveFrame* frame);

/** \brief Primitive. Add the operand to self. */
void SmallInteger_primAdd(PrimitiveFrame* frame);

/** \brief Primitive. Subtract the operand from self. */
void SmallInteger_primSub(PrimitiveFrame* frame);

/** \brief Primitive. Multiplicate the operand and self. */
void SmallInteger_primMul(PrimitiveFrame* frame);

/** \brief Primitive. Divide self by the operand and round down. */
void SmallInteger_primDivFloor(PrimitiveFrame* frame);

/** \brief Primitive. Return the result of self modulo the operand. */
void SmallInteger_primMod(PrimitiveFrame* frame);

/** \brief Primitive. Check, whether self is less than the operand. */
void SmallInteger_primLessThan(PrimitiveFrame* frame);

/** \brief Primitive. Check, whether self is less or equal to the operand. */
void SmallInteger_primLessThanOrEqual(PrimitiveFrame* frame);

/** \brief Primitive. Return the result of self AND the operand. */
void SmallInteger_primBitAnd(PrimitiveFrame* frame);

/** \brief Primitive. Return the result of self OR the operand. */
void SmallInteger_primBitOr(PrimitiveFrame* frame);

/** \brief Primitive. Return the bit result of self XOR the operand. */
void SmallInteger_primBitXor(PrimitiveFrame* frame);

/** \brief Primitive. Shift self left or right by operand bits. */
void SmallInteger_primBitShift(PrimitiveFrame* frame);

/** \brief Primitive. Return a random number between 1 and self. */
void SmallInteger_primAtRandom(PrimitiveFrame* frame);

/** \brief Primitive. Add the operand to self. */
void LargeInteger_primAdd(PrimitiveFrame* frame);

/** \brief Primitive. Subtract the operand from self. */
void LargeInteger_primSub(PrimitiveFrame* frame);

/** \brief Primitive. Multiplicate self with the operand. */
void LargeInteger_primMul(PrimitiveFrame* frame);

/** \brief Primitive. Print the LargeInteger. */
void LargeInteger_primPrint(PrimitiveFrame* frame);

/** \brief Primitive. Check, whether self is less than the operand. */
void LargeInteger_primLessThan(PrimitiveFrame* frame);

/** \brief Primitive. Check, whether self and operad have an equal value. */
void LargeInteger_primEquals(PrimitiveFrame* frame);
/** @} */

#endif // _INTEGER_H_
