/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

#include "interpreter.h"

#include "object.h"
#include "method.h"
#include "class.h"
#include "motor.h"
#include "frame.h"
#include "integer.h"
#include "display.h"
#include "process.h"
#include "scheduler.h"
#include "primitives.h"
#include "garbagecollection.h"
#include "methodcache.h"
#include "sensors.h"
#include "communication.h"
#include "system.h"

FrameRef currentFrame;
Frame* currentFramePtr;

U8* bytecodeStart;
U8* instructionPointer;
ObjectRef currentReceiver;
const Method* currentMethodPtr;

ProcessRef currentProcessRef;
Process* currentProcess;
S16 currentStackTopIndex;
ObjectRef* currentStack;

#define SENDS_TO_SCHEDULE 100
int sendsToSchedule;

ObjectRef Interpreter_stack() {
  return currentProcess->stack;
}

static ObjectRef Interpreter_stackTop() {
  return currentStack[currentStackTopIndex];
}

void Interpreter_shutdown(bool save) {
  if (save) {
    NXT_printString("Shutting down... ");
    // Remove all process references from the scheduler
    Scheduler* scheduler = Scheduler_getForWrite(specialObjects->Processor);
    scheduler->firstLink = specialObjects->nil;
    scheduler->lastLink = specialObjects->nil;
    scheduler->schedulerProcess = specialObjects->nil;
    scheduler->activeProcess = specialObjects->nil;
    scheduler->waitingProcesses = specialObjects->nil;
    // Garbage collect everything and compact to Flash
    GarbageCollection_finalCollect();
    // Write object table
    Image_storeObjectTable();
    // Update object count in header
    Image_updateHeader();
  }
  NXT_shutdown();
}

void Interpreter_loadInstructionStream() {
  if (currentFrame != specialObjects->nil) {
    currentFramePtr = Frame_get(currentFrame);
    currentMethodPtr = Frame_getMethod(currentFramePtr);
    bytecodeStart = Method_getBytecodeStart(currentMethodPtr);
    instructionPointer =  bytecodeStart + Frame_getPC(currentFramePtr);
    currentReceiver = Frame_getReceiver(currentFramePtr);
  } else {
    ProcessRef schedulerProcess = Scheduler_get(specialObjects->Processor)->schedulerProcess;
    if (schedulerProcess == currentProcessRef) {
      // Scheduler terminates, shutdown whole machine
      Interpreter_shutdown(FALSE);
    }
    Interpreter_switchToProcess(schedulerProcess);
  }
}

void Interpreter_storeInstructionStream() {
  if (currentFrame != specialObjects->nil) {
    Frame_setPC(currentFramePtr, instructionPointer - bytecodeStart);
  }
}

void Interpreter_storeExecutionState() {
  Interpreter_storeInstructionStream();
  Process_setStackTopIndex(Process_getForWrite(currentProcessRef), currentStackTopIndex);
}

void Interpreter_reloadExecutionState() {
  // load process and frame information from current process
  Process* process = Process_getForWrite(currentProcessRef);
  currentProcess = process;
  currentStack = IndexableObject_get(process->stack)->instanceVariables;
  currentStackTopIndex = Process_getStackTopIndex(process);
  currentFrame = currentProcess->currentFrame;
  Interpreter_loadInstructionStream();
}

static int Interpreter_calcBytecodeIndex(U8 bytecode) {
  int index = bytecode >> 4;
  if (index == 0) {
    return bytecode;
  } else {
    return index + 16;
  }
}

void Interpreter_run(void) {
  static void* jumpTable[] = {
    &&Return, &&Dup, &&Pop, &&PushSelf, &&PushNil, &&PushTrue, &&PushFalse, &&PushThisContext,
    &&PushArgument, &&PushArgument, &&PushArgument, &&PushArgument, &&PushArgument, &&PushArgument, &&PushArgument, &&NonLocalReturn,

    &&PushTemporary, &&PushTemporary, &&PushInstanceVariable, &&PushInstanceVariable, &&PopInstanceVariable, &&PopInstanceVariable, &&PushLiteral, &&PushLiteral,
    &&PushBlock, &&PushBlock, &&SuperSend, &&SuperSend, &&Send, &&Send, &&JumpIfTrue, &&JumpBack};

  U8 bytecode;
  goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];

  NonLocalReturn: {
    NXT_debugString("NON LOCAL RETURN\n");
    Interpreter_nonLocalReturn();
    Interpreter_loadInstructionStream();
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  Return: {
    NXT_debugString("RET\n");
    Interpreter_return();
    Interpreter_loadInstructionStream();
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  Dup: {
    NXT_debugString("DUP, ");
    ObjectRef dupped = Interpreter_stackTop();
    Interpreter_stackPush(dupped);
    Object_incReferenceCount(dupped);
    NXT_debugString("Stack top = ");
    NXT_debugNumber(Interpreter_stackTopIndex());
    NXT_debugString("\n");
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  Pop: {
    NXT_debugString("POP, ");
    ObjectRef object = Interpreter_stackPop();
    Object_decReferenceCount(object);
    NXT_debugString("Stack top = ");
    NXT_debugNumber(Interpreter_stackTopIndex());
    NXT_debugString("\n");
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  PushSelf: {
    NXT_debugString("PUSH SELF\n");
    ObjectRef self = currentReceiver;
    Object_incReferenceCount(self);
    Interpreter_stackPush(self);
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  PushNil: {
    NXT_debugString("PUSH NIL\n");
    Interpreter_stackPush(specialObjects->nil);
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  PushTrue: {
    NXT_debugString("PUSH TRUE\n");
    Interpreter_stackPush(specialObjects->true);
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  PushFalse: {
    NXT_debugString("PUSH FALSE\n");
    Interpreter_stackPush(specialObjects->false);
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  PushThisContext: {
    NXT_debugString("PUSH THIS_CONTEXT\n");
    Object_incReferenceCount(currentFrame);
    Interpreter_stackPush(currentFrame);
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  PushArgument: {
    NXT_debugString("PUSH ARG ");
    NXT_debugNumber(bytecode & 0x07);
    ObjectRef argument = Frame_getArgument(currentFramePtr, bytecode & 0x07);
    NXT_debugString(" = ");
    NXT_debugNumber(argument);
    NXT_debugString("\n");
    Object_incReferenceCount(argument);
    Interpreter_stackPush(argument);
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  PushTemporary: {
    if ((bytecode & 0x08) != 0) {
      goto PopTemporary;
    }
    NXT_debugString("PUSH TEMP ");
    NXT_debugNumber(bytecode & 0x07);
    ObjectRef temp = Frame_getTemporary(currentFramePtr, bytecode & 0x07);
    NXT_debugString(" = ");
    NXT_debugNumber(temp);
    NXT_debugString("\n");
    Object_incReferenceCount(temp);
    Interpreter_stackPush(temp);
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  PopTemporary: {
    NXT_debugString("POP into TEMP ");
    NXT_debugNumber(bytecode & 0x07);
    ObjectRef value = Interpreter_stackPop();
    NXT_debugString(" = ");
    NXT_debugNumber(value);
    NXT_debugString("\n");
    ObjectRef oldValue = Frame_setTemporary(currentFramePtr, bytecode & 0x07, value);
    Object_decReferenceCount(oldValue);
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  PushInstanceVariable: {
    NXT_debugString("PUSH INST ");
    NXT_debugNumber(bytecode & 0x1F);
    ObjectRef instVar = Object_getInstVar(currentReceiver, bytecode & 0x1F);
    NXT_debugString(" = ");
    NXT_debugNumber(instVar);
    NXT_debugString("\n");
    Object_incReferenceCount(instVar);
    Interpreter_stackPush(instVar);
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  PopInstanceVariable: {
    NXT_debugString("POP into INST");
    NXT_debugNumber(bytecode & 0x1F);
    ObjectRef instVar = Interpreter_stackPop();
    NXT_debugString(" = ");
    NXT_debugNumber(instVar);
    NXT_debugString("\n");
    // We have an object, which has typical instance variables and thus is not indexable
    Object_setInstVar(currentReceiver, bytecode & 0x1F, instVar);
    Object_decReferenceCount(instVar);
    //ObjectMemory_getForWrite(currentReceiver)->instanceVariables2[bytecode & 0x1F] = instVar;
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  PushLiteral: {
    NXT_debugString("PUSH LITERAL ");
    NXT_debugNumber(bytecode & 0x1F);
    ObjectRef object = Method_getLiteral(currentMethodPtr, bytecode & 0x1F);
    NXT_debugString(" = ");
    NXT_debugNumber(object);
    Object_incReferenceCount(object);
    Interpreter_stackPush(object);
    NXT_debugString(", Stack top = ");
    NXT_debugNumber(Interpreter_stackTopIndex());
    NXT_debugString("\n");
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  PushBlock: {
    NXT_debugString("PUSH BLOCK from ");
    ObjectRef object = Method_getLiteral(currentMethodPtr, bytecode & 0x1F);
    NXT_ASSERT(!ObjectTable_isFree(object), "Block delete!???");
    NXT_debugString(": Object = ");
    NXT_debugNumber(object);
    FrameRef blockFrame = Frame_newBlock(object, currentFrame);
    Object_incReferenceCount(blockFrame);
    Interpreter_stackPush(blockFrame);
    NXT_debugString(", Stack top = ");
    NXT_debugNumber(Interpreter_stackTopIndex());
    NXT_debugString("\n");
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  SuperSend: {
    ObjectRef symbol = Method_getLiteral(currentMethodPtr, bytecode & 0x1F);
    ObjectRef receiver = Interpreter_stackAt(Interpreter_stackTopIndex() - Symbol_countArguments(symbol));
    NXT_debugString("SUPER_SEND to ");
    NXT_debugNumber(receiver);
    NXT_debugString(" with #");
    if (DEBUG > 1) String_print(symbol);
    NXT_debugString("\n");
    Interpreter_send(receiver, symbol, TRUE);
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  Send: {
    ObjectRef symbol = Method_getLiteral(currentMethodPtr, bytecode & 0x1F);
    ObjectRef receiver = Interpreter_stackAt(Interpreter_stackTopIndex() - Symbol_countArguments(symbol));
    NXT_debugString("SEND to ");
    NXT_debugNumber(receiver);
    NXT_debugString(" with #");
    if (DEBUG > 1) String_print(symbol);
    NXT_debugString("\n");
    Interpreter_send(receiver, symbol, FALSE);
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  JumpIfTrue: {
    NXT_debugString("JUMP IF TRUE: ");
    ObjectRef object = Interpreter_stackPop();
    if (object == specialObjects->true) {
      instructionPointer += (bytecode & 0x0F) + 1;
      NXT_debugNumber((bytecode & 0x0F) + 1);
      NXT_debugString("\n");
    } else {
      NXT_debugNumber(0);
      NXT_debugString("\n");
    } // TODO: Throw error, if no boolean is found
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
  JumpBack: {
    NXT_debugString("JUMP BACK: ");
    ObjectRef object = Interpreter_stackPop();
    Object_decReferenceCount(object);
    instructionPointer -= (bytecode & 0x0F) + 1 + 1;
    NXT_debugNumber((bytecode & 0x0F) + 1 + 1);
    NXT_debugString("\n");
    goto *jumpTable[Interpreter_calcBytecodeIndex(bytecode = *instructionPointer++)];
  }
}

void Interpreter_printStackTrace() {
  FrameRef frame = currentFrame;
  while (frame != specialObjects->nil) {
    Frame* framePtr = Frame_get(frame);
    ClassRef class = Object_getClass(Frame_getReceiver(framePtr));
    String_print(Class_get(class)->name);
    if (!Frame_executingBlock(framePtr)) {
      if (MethodFrame_getOwner(framePtr) != class) {
        NXT_printString("(");
        String_print(Class_get(MethodFrame_getOwner(framePtr))->name);
        NXT_printString(")");
      }
      NXT_printString(">>#");
      String_print(Method_getSelector(Frame_getMethod(framePtr)));
    } else {
      NXT_printString(">>[]");
    }
    NXT_printString(", ");
    frame = framePtr->sender;
  }
}

void Interpreter_send(ObjectRef receiver, ObjectRef symbol, bool super_send) {
  MethodRef methodRef;
  ClassRef lookupClass;
  if (!super_send) {
    lookupClass = Object_getClass(receiver);
  } else {
    lookupClass = Class_get(MethodFrame_getOwner(currentFramePtr))->superclass;
  }
  methodRef = MethodCache_lookup(lookupClass, symbol);
  if (methodRef == specialObjects->nil) {
    //methodRef = Class_lookupMethod(lookupClass, symbol);
    // Method not found, send DNU
    int argumentCount = Symbol_countArguments(symbol);
    ObjectRef arguments = Class_newWithSize(
        specialObjects->NXTArray, argumentCount);
    for (int i = 0; i < argumentCount; i++) {
      ObjectRef argument = Interpreter_stackPop();
      Object_setInstVar(arguments, i, argument);
      Object_decReferenceCount(argument);
    }
    ObjectRef message = Class_new(specialObjects->NXTMessage);
    Object_setInstVar(message, 0, symbol);
    Object_setInstVar(message, 1, arguments);
    Object_setInstVar(message, 2, lookupClass);
    Interpreter_stackPush(message);
    Object_incReferenceCount(message);
    Interpreter_send(receiver, specialObjects->doesNotUnderstand, FALSE);
    return;
  }
  NXT_ASSERT(!ObjectTable_isFree(methodRef), "Corrupt method");
  Method* method = Method_get(methodRef);
  if (Method_isPrimitive(method)) {
    // we got a primitive
    U16 primitive = Method_getPrimitiveNumber(method);
    if (primitive >= PRIMITIVE_COUNT || primitives[primitive] == NULL) {
      NXT_printString("Unknown primitive ");
      NXT_printNumber(primitive);
      NXT_error("Halting...");
    }
    // Call primitive
    PrimitiveFrame primitiveFrame;
    primitiveFrame.self = receiver;
    primitiveFrame.argumentCount = Method_getArgumentCount(method);
    primitives[primitive](&primitiveFrame);
    Primitive_emptyTemporarySet();
  } else {
    U8 argumentCount = Method_getArgumentCount(method);
    Interpreter_storeInstructionStream();
    Interpreter_executeFrame(Frame_newMethod(receiver, methodRef), argumentCount);
    Interpreter_loadInstructionStream();
    if (!schedulerRunning) {
      sendsToSchedule--;
    }
    if (schedulingEnabled && sendsToSchedule == 0) {
      Interpreter_switchToProcess(Scheduler_get(specialObjects->Processor)->schedulerProcess);
      schedulerRunning = TRUE;
    }
  }
}

void Interpreter_initialize() {
  MethodCache_reset();

  currentProcessRef = Process_new();
  currentProcess = Process_getForWrite(currentProcessRef);
  Scheduler_setSchedulerProcess(Scheduler_getForWrite(specialObjects->Processor), currentProcessRef);
  Scheduler_setActiveProcess(Scheduler_getForWrite(specialObjects->Processor), currentProcessRef);
  currentFrame = specialObjects->nil;
  currentStack = IndexableObject_get(Process_get(currentProcessRef)->stack)->instanceVariables;
  currentStackTopIndex = Process_getStackTopIndex(Process_get(currentProcessRef));
  sendsToSchedule = SENDS_TO_SCHEDULE;
  schedulerRunning = TRUE;
  schedulingEnabled = TRUE;
}

void Interpreter_switchToProcess(ProcessRef newProcess) {
  schedulerRunning = FALSE;
  sendsToSchedule = SENDS_TO_SCHEDULE;
  Process* currentProcessPtr = Process_getForWrite(currentProcessRef);
  // store old process and frame information
  Interpreter_storeInstructionStream();
  Process_setStackTopIndex(currentProcessPtr, currentStackTopIndex);
  // load process and frame information from new process
  currentProcessRef = newProcess;
  Interpreter_reloadExecutionState();
}

void Interpreter_return(void) {
  Frame* self = Frame_get(currentFrame);
  FrameRef newFrame = self->sender;

  // Remove all temporaries, arguments and intermediate values from stack but preserve return value
  ObjectRef returnValue = Interpreter_stackTop();
  Interpreter_stackPop();

  self->sender = specialObjects->nil;

  S16 oldStackPointer = Interpreter_stackTopIndex();
  S16 newStackPointer = SmallInteger_fromReference(self->stackPointer) - 1 - 1;

  // Decrement references on self and all immediate values on stack
  int startIndex = newStackPointer + 1;
  ObjectRef nil = specialObjects->nil;
  for (int i = startIndex; i <= oldStackPointer; i++) {
    ObjectRef stackValue = Interpreter_stackAt(i);
    if (stackValue != nil) {
      Object_decReferenceCount(stackValue);
      // Nil stack for testing purposes
      Interpreter_stackAtPut(i, nil);
    }
  }
  // Also, decrement reference count on left frame
  Object_decReferenceCount(currentFrame);

  Interpreter_stackSetTopIndex(newStackPointer);
  Interpreter_stackPush(returnValue);

  currentFrame = newFrame;
  Process_setCurrentFrame(currentProcess, newFrame);
}

void Interpreter_nonLocalReturn(void) {
  Frame* self = Frame_get(currentFrame);
  if (Frame_executingBlock(self)) {
    FrameRef homeMethodFrame = Frame_getHomeMethodFrame(self);
    while (currentFrame != specialObjects->nil && currentFrame != homeMethodFrame) {
      self = Frame_get(currentFrame);
      if (!Frame_executingBlock(self) && Method_getSelector(Frame_getMethod(self)) == specialObjects->onNonLocalReturnsDo) {
        // Catched non local return, execute handling block
        // on top of the stack, we have the return value. however, we need the block before, to do a regular aBlock>>value
        ObjectRef returnValue = Interpreter_stackPop();
        FrameRef handler = Frame_getArgument(self, 0);
        Object_incReferenceCount(handler);
        Interpreter_stackPush(handler);
        Interpreter_stackPush(returnValue);
        U8 argumentCount = Method_getArgumentCount(Frame_getMethod(self));
        Interpreter_executeFrame(handler, argumentCount);
      }
      Interpreter_return();
    }
    if (currentFrame == specialObjects->nil) {
      NXT_error("Non-local return in block, which has left its context!");
    }
  }
  Interpreter_return();
}

void Interpreter_executeFrame(FrameRef frame, U8 argumentCount) {
  Frame* self = Frame_get(frame);
  // No need to increment RC of context, because context/currentFrame will lose one reference
  NXT_ASSERT(self->sender == specialObjects->nil, "check");
  self->sender = currentFrame;
  Frame_setPC(self, 1);
  // inc reference count on frame
  Object_incReferenceCount(frame);
  U16 tempCount = Frame_getTempCount(self);
  ObjectRef nil = specialObjects->nil;
  // move arguments from stack to frame
  for (int i = argumentCount - 1; i >= 0; i--) {
    self->argsAndTemps[i] = Interpreter_stackPop();
  }
  // nil temps, if necessary
  for (int i = 0; i < tempCount; i++) {
    if (self->argsAndTemps[argumentCount + i] != nil) {
      Object_decReferenceCount(self->argsAndTemps[argumentCount + i]);
      self->argsAndTemps[argumentCount + i] = nil;
    }
  }
  // store the current stack pointer
  Frame_setStackPointer(self, Interpreter_stackTopIndex());
  currentFrame = frame;
  Process_setCurrentFrame(currentProcess, frame);
}
