/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

/** \file interpreter.h
 * \author Martin Beck
 * \brief Functions for the interpreter.
 */

#ifndef _INTERPRETER_H_
#define _INTERPRETER_H_

#include "object.h"
#include "method.h"
#include "frame.h"
#include "process.h"

/** \brief A reference to the currently executed frame. */
extern FrameRef currentFrame;

/** \brief Current top of the stack. */
extern S16 currentStackTopIndex;

/** \brief Direct pointer to the current stack. */
extern ObjectRef* currentStack;


/** \brief Initialize the interpreter. */
void Interpreter_initialize(void);

/** \brief Start the interpreter. Does not return. */
void Interpreter_run(void);

/** \brief Sends a message.
 *
 * Directs the interpreter to send a message to the given object
 * in the current context. The interpreter changes its state
 * accordingly to continue execution in the invoked method.
 *
 * \param receiver  The object which receives the message
 * \param message   The message which is sent.
 * \param superSend Whether the send is super send
 */
void Interpreter_send(ObjectRef receiver, ObjectRef message, bool superSend);

/** \brief Prints a Smalltalk stack trace for debugging purposes. */
void Interpreter_printStackTrace(void);

/** \brief Switches to another process.
 *
 * The interpreter stores its current process data in the corresponding
 * Smalltalk objects and loads its new state from the process to switch
 * to.
 *
 * \param newProcess The process, the interpreter should switch to.
 */
void Interpreter_switchToProcess(ProcessRef newProcess);

/** \brief Directs the interpreter to a new frame.
 *
 * The interpreter connects the new frame with the current frame as sender
 * and fills the new frame with the arguments from the stack. Afterwards
 * the current frame is set to the new one and the according data and
 * bytecode pointers are loaded.
 *
 * \param frame         The Frame to execute.
 * \param argumentCount The number of arguments to move from stack to Frame.
 */
void Interpreter_executeFrame(FrameRef frame, U8 argumentCount);

/** \brief Directs the interpreter to return one frame in the sender chain. */
void Interpreter_return(void);

/** \brief Does a non-local return from a block. */
void Interpreter_nonLocalReturn(void);

/** \brief Returns a reference to the current process' stack. */
ObjectRef Interpreter_stack(void);

/** \brief Loads the interpreter state for the current frame. */
void Interpreter_loadInstructionStream(void);

/** \brief Stores the current interpreter state in the current frame object. */
void Interpreter_storeInstructionStream(void);

/** \brief Stores the current process state. */
void Interpreter_storeExecutionState(void);

/** \brief Reloads the current execution state from the process objects. */
void Interpreter_reloadExecutionState(void);

/** \brief Pushes a value on the current stack.
 *
 * \param object Value to push.
 */
static inline void Interpreter_stackPush(ObjectRef object) {
  NXT_ASSERT(currentStackTopIndex < STACK_SIZE, "Stack overflow");
  currentStack[++currentStackTopIndex] = object;
}

/** \brief Pop a value from the current stack.
 *
 * \return The popped value.
 */
static inline ObjectRef Interpreter_stackPop() {
  NXT_ASSERT(currentStackTopIndex > -1, "Stack underflow");
  ObjectRef object = currentStack[currentStackTopIndex--];
  return object;
}

/** \brief Returns the current stack top index.
 *
 * \return Current index of the stack topmost value.
 */
static inline S16 Interpreter_stackTopIndex() {
  return currentStackTopIndex;
}

/** \brief Read the stack at given index.
 *
 * \param index The index to the current stack.
 * \return The value from the stack, without popping anything.
 */
static inline ObjectRef Interpreter_stackAt(S16 index) {
  return currentStack[index];
}

/** \brief Modify the stack at a given index.
 *
 * \param index The index to the current stack.
 * \param value The new value.
 */
static inline void Interpreter_stackAtPut(S16 index, ObjectRef value) {
  currentStack[index] = value;
}

/** \brief Sets the current stack top index.
 *
 * \param index The new index of the stack's topmost value.
 */
static inline void Interpreter_stackSetTopIndex(S16 index) {
  currentStackTopIndex = index;
}

/** \brief Shuts the interpreter and the system down.
 *
 * This function shuts the whole system down. It is also
 * able to store the important parts of the image back to
 * Flash. However, it still deletes all processes and
 * everything which is only referenced by them.
 *
 * \param save If TRUE, all live objects except processes
 *             are written back to Flash.
 */
void Interpreter_shutdown(bool save);

#endif // _INTERPRETER_H_
