/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

 /** \file method.h
  * \author Martin Beck
  * \brief Functions for compiled methods and blocks.
  */

#ifndef _METHOD_H_
#define _METHOD_H_

#include "object.h"
#include "integer.h"

/** \brief Type redefinition to indicate Method object references. */
typedef ObjectRef MethodRef;

/** \brief Type redefinition to indicate Block object references. */
typedef MethodRef BlockRef;

/** \brief Method representation in memory.
 *
 * A compiled method is a special object in terms of the contained
 * elements. Although the class format indicates, that the element
 * size of instances is one byte, it nevertheless contains several regular
 * references to other objects. In fact, a method object starts with a
 * two-byte header and multiple references to the literals used in that
 * method. The last two literals always reference the selector and the
 * class of the method object. Afterwards, the real bytecodes are stored
 * one by one.
 */
struct Method {

  /** \brief Class reference. */
  ObjectRef klass;

  /** \brief Size of the following data in bytes. */
  U16 size;

  /** \brief Header for this method.
   *
   * The header of a method contains several important data and is encoded
   * as a tagged-integer. The three lowest bits represent the number of
   * temporary variables used by this method. Afterwards, 5 bits encode the
   * number of literals. Next, one bit is used to indicate, whether this
   * method object encapsulates a primitive. If yes, the 8 lowest bits
   * represent the primitive number, instead. The next 3 bits are always used
   * for the argument count.
   */
  SmallInteger header;

  /** \brief References to the literals used by this method. */
  ObjectRef literals[];
} __attribute__ ((packed));
typedef struct Method Method;

/** \brief Compiled blocks have exactly the same representation as methods. */
typedef Method Block;

/** \brief Resolve a Method reference.
 *
 * Retrieves a direct pointer to a Method object.
 *
 * \param methodRef The reference to the method.
 * \return The direct pointer to the Method object in memory.
 */
static inline Method* Method_get(MethodRef methodRef) {
  return (Method*)ObjectTable_getAddress(methodRef);
}

/** \brief Resolve a Method reference for writing.
 *
 * Retrieves a direct pointer to a Method object ensuring it lies entirely
 * within RAM.
 *
 * \param methodRef The reference to the method.
 * \return The direct pointer to the Method object in RAM.
 */
static inline Method* Method_getForWrite(MethodRef methodRef) {
  return (Method*)Object_getForWrite(methodRef);
}

/** \brief Checks, whether the given object is a real method.
 *
 * \param self An object.
 * \return TRUE, if it is a method, FALSE if not.
 */
bool Method_isMethod(const Method* self);

/** \name Getters
 * @{
 */

/** \brief Returns the method header.
 *
 * \param self The method.
 * \return The method header of the given method.
 */
static inline U16 Method_getHeader(const Method* self) {
  return SmallInteger_fromReference(self->header);
}

/** \brief Returns the number of temporary variables.
 *
 * \param self The method.
 * \return The number of temporary variables used by the given method.
 */
static inline U8 Method_getTempCount(const Method* self) {
  return Method_getHeader(self) & 0x07;
}

/** \brief Returns the number of arguments.
 *
 * \param self The method.
 * \return The number of arguments the given method expects.
 */
static inline U8 Method_getArgumentCount(const Method* self) {
  return Method_getHeader(self) >> 9 & 0x07;
}

/** \brief Returns the number literals.
 *
 * \param self The method.
 * \return The number of literals used and referenced by the given method.
 */
static inline U8 Method_getLiteralCount(const Method* self) {
  return Method_getHeader(self) >> 3 & 0x1F;
}

/** \brief Returns the address of the first bytecode.
 *
 * \param self The method.
 * \return A pointer to the first bytecode of the given method.
 */
static inline U8* Method_getBytecodeStart(const Method* self) {
  return ((U8*) &self->literals) + Method_getLiteralCount(self) * 2;
}

/** \brief Retrieves a literal reference of the method.
 *
 * \param self          The method.
 * \param literalNumber Number of the literal to fetch.
 * \return The literal reference of the method specified by the given number.
 */
static inline ObjectRef Method_getLiteral(const Method* self, U8 literalNumber) {
  return self->literals[literalNumber];
}

/** \brief Indicates, whether the method encapsulates a primitive.
 *
 * \param self The method.
 * \return TRUE, if the given method is a primitive, FALSE if not.
 */
static inline bool Method_isPrimitive(const Method* self) {
  return Method_getHeader(self) >> 8 & 1;
}

/** \brief Returns the primitive number.
 *
 * \param self The method.
 * \return  The primitive number of the given method. Undefined, if the method
 *          is no primitive.
 */
static inline U16 Method_getPrimitiveNumber(const Method* self) {
  return Method_getHeader(self) & 0xFF;
}

/** \brief Returns the class, where the method is defined in.
 *
 * \param self The method.
 * \return A reference to the class, which the contains the given method.
 */
static inline ObjectRef Method_getMethodClass(const Method* self) {
  return self->literals[Method_getLiteralCount(self) - 1];
}

/** \brief Returns the selector under which the method is stored.
 *
 * \param self The method.
 * \return A reference to the symbol, which is the name of the given method.
 */
static inline ObjectRef Method_getSelector(const Method* self) {
  return self->literals[Method_getLiteralCount(self) - 2];
}
/** @} */

/** \name Primitives
 * @{
 */

/** \brief Allows access to the method header and the literals from Smalltalk.
 *
 * \param frame Contains necessary context information for the primitive.
 */
void Method_primObjectAt(PrimitiveFrame* frame);

/** \brief Allows writing access to the method header and the literals.
 *
 * \param frame Contains necessary context information for the primitive.
 */
void Method_primObjectAtPut(PrimitiveFrame* frame);
/** @} */

#endif // _METHOD_H_
