/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

#ifndef _NXT_API_H_
#define _NXT_API_H_

#ifndef NXT_NO_TYPE_DEFINTIONS
typedef unsigned char U8;
typedef signed char S8;
typedef unsigned short U16;
typedef signed short S16;
typedef unsigned int U32;
typedef signed int S32;

typedef U8 bool;

#define FALSE (0)
#define TRUE (!FALSE)

#ifndef NULL
#define NULL ((void*)0)
#endif

#define MIN(A, B) (A < B ? A : B)
#define MAX(A, B) (A > B ? A : B)

typedef enum {
  BUTTON_NONE = 0,
  BUTTON_OK,
  BUTTON_CANCEL,
  BUTTON_LEFT,
  BUTTON_RIGHT
} NXTButton;
#endif

#define UNUSED(x) x __attribute__ ((unused))

typedef void (*NXTClosure)(void);

/** \name Display functions
 * @{
 */

/** \brief Initializes the display. */
void NXTDisplay_init(void);

/** \brief Shuts the display module down. */
void NXTDisplay_shutdown(void);

/** \brief Clears the display. */
void NXTDisplay_clear(void);

/** \brief Print a single character.
 *
 * Prints the given character to the current display position.
 *
 * \param character The character to print.
 */
void NXT_printChar(char character);

/** \brief Print a string.
 *
 * Prints the given NULL-terminated string to the current display position.
 *
 * \param string The string to print.
 */
void NXT_printString(const char* string);

/** \brief Print a number.
 *
 * Prints the given number to the current display position.
 *
 * \param number The unsigned integer to print.
 */
void NXT_printNumber(U32 number);

/** \brief Print a number in hex format.
 *
 * Prints the given number to the current display position in hex format.
 *
 * \param number The unsigned integer to print.
 */
void NXT_printHexNumber(U32 number);

/** \brief Sets the scrolling mode.
 *
 * \param on If TRUE, the display scrolls, if printing leaves the screen
 *           at the bottom. If FALSE, the text will be wrapped to the top
 *           of the display.
 */
void NXTDisplay_setScrollMode(bool on);

/** \brief Returns the cursor column.
 *
 * \return X position of the cursor.
 */
U8 NXTDisplay_getCursorX(void);

/** \brief Returns the cursor row.
 *
 * \return Y position of the cursor.
 */
U8 NXTDisplay_getCursorY(void);

/** \brief Sets the cursor to a new position.
 *
 * Cursor positions are expressed in lines and columns, where the size
 * of each is exactly one character.
 *
 * \param x Column where to put subsequent text.
 * \param y Row where to put subsequent text.
 */
void NXTDisplay_setCursor(U8 x, U8 y);
/** @} */

/** \name USB functions
 * @{
 */

/** \brief Initialize the usb module. */
void NXTUsb_init(void);

/** \brief Shuts the usb module down. */
void NXTUsb_shutdown(void);

/** \brief Indicates a present USB connection.
 *
 * \return Whether the NXT has a USB connection or not.
 */
U8 NXTUsb_isConnected(void);

/** \brief Returns the USB maximum packet size.
 *
 * \return Maximum size of an USB packet in bytes.
 */
U32 NXTUsb_packetSize(void);

/** \brief Read data from the USB port asynchroneously.
 *
 * Specifies, where the next USB packet data will be stored. If a packet
 * is smaller than the provided buffer, not the whole buffer is used.
 *
 * \param data   A pointer to the buffer.
 * \param length The length of the buffer in bytes.
 */
void NXTUsb_read(U8* data, U32 length);

/** \brief Indicates, when data has been read.
 *
 * \return The amount of bytes read with the last packet. Reset to 0 after
 *         each call.
 */
U32 NXTUsb_dataRead(void);

/** \brief Writes data to the USB port asynchroneously.
 *
 * \param data   The data to write.
 * \param length Length of the data.
 */
void NXTUsb_write(U8* data, U32 length);

/** \brief Returns, whether the last data packet has been written.
 *
 * \return TRUE, if packet has been sent. FALSE, if not.
 */
bool NXTUsb_dataWritten(void);

/** \brief Reads data from the USB synchroneously.
 *
 * \param data   Buffer for the read data.
 * \param length Size of the buffer in bytes.
 */
void NXTUsb_readSync(U8* data, U32 size);
/** @} */

/** \name Memory functions
 * @{
 */

/** \brief Initializes the RAM module. */
void NXTMemory_init(void);

/** \brief Shuts the RAM module down. */
void NXTMemory_shutdown(void);

/** \brief Gets a pointer to the free RAM start.
 *
 * Returns the starting address of the RAM region, which can be freely used
 * by the VM.
 *
 * \return Address of the free RAM region.
 */
U8* NXTMemory_getStart(void);

/** \brief Gets a pointer to the end of the free RAM.
 *
 * Returns the first addres, which is NOT in the RAM region freely usable
 * by the VM.
 *
 * \return Address of the first byte after the free RAM region.
 */
U8* NXTMemory_getEnd(void);

/** \brief Returns the size of the freely usable RAM.
 *
 * \return Size of the freely usable RAM region.
 */
U32 NXTMemory_getSize(void);

/** \brief Indicates, whether an address is within RAM.
 *
 * \param address The address to check.
 * \return TRUE, if the given address lies within the RAM.
 */
bool NXTMemory_isRAMAddress(void* address);

/** \brief Copies a section in memory to another one.
 *
 * \param dest   The destination of the copying.
 * \param source The source of the copy process.
 * \param length The number of bytes to copy.
 */
void NXTMemory_copy(void* dest, void* source, U32 length);

/** \brief Moves a section in memory to another one.
 *
 * This operations ensures that even overlapping source
 * and destination addresses are copied correctly.
 *
 * \param dest   The destination of the moving.
 * \param source The source of the move process.
 * \param length The number of bytes to move.
 */
void NXTMemory_move(void* dest, void* source, U32 length);
/** @} */

/** \name Flash functions
 * @{
 */

/** \brief Initialize Flash module. */
void NXTFlash_init(void);

/** \brief Shuts the Flash module down. */
void NXTFlash_shutdown(void);

/** \brief Returns the address of the freely usable Flash region.
 *
 * \return Address of the freely usable Flash region.
 */
U8* NXTFlash_getStart(void);

/** \brief Returns the end of the free Flash.
 *
 * \return Address of the first byte behind the freely usable Flash
 *         region.
 */
U8* NXTFlash_getEnd(void);

/** \brief Returns the size of the free Flash region.
 *
 * \return Size of the freely usable Flash region in bytes.
 */
U32 NXTFlash_getSize(void);

/** \brief Returns the size of one page in the Flash memory.
 *
 * \return Size of one page in the Flash memory.
 */
U32 NXTFlash_getPageSize(void);

/** \brief Reads a page from Flash memory.
 *
 * \param page Number of the page to read.
 * \param data Buffer to hold the data of the page.
 */
void NXTFlash_readPage(U32 page, U8* data);

/** \brief Write a page to the Flash memory.
 *
 * \param data The data to write.
 * \param page The number of the page to write.
 */
void NXTFlash_writePage(U8* data, U32 page);

/** \brief Returns the number of the first page pointing to the free Flash.
 *
 * \return Returns the number of that page, where the freely usable Flash
 *         memory begins.
 */
U32 NXTFlash_firstPage(void);

/** \brief Returns the number of the last page in free Flash.
 *
 * \return Number of the last page in freely usable Flash memory.
 */
U32 NXTFlash_lastPage(void);
/** @} */

/** \name Bluetooth functions
 * @{
 */
void NXTBluetooth_init(void);
void NXTBluetooth_shutdown(void);
void NXTBluetooth_setDiscoverable(bool on);
bool NXTBluetooth_isConnectionPending(void);
bool NXTBluetooth_isWaitingForPIN(void);
void NXTBluetooth_sendPin(char* pin);
void NXTBluetooth_acceptConnection(bool accept);
S32 NXTBluetooth_connectionEstablished(void);
void NXTBluetooth_openStream(S32 handle);
void NXTBluetooth_closeStream(void);
bool NXTBluetooth_isStreamOpen(void);
void NXTBluetooth_read(U8* data, U32 length);
U32 NXTBluetooth_dataRead(void);
void NXTBluetooth_write(U8* data, U32 length);
bool NXTBluetooth_dataWritten(void);

/** \brief Reads data from Bluetooth synchroneously.
 *
 * \param data   Buffer for the read data.
 * \param length Size of the buffer in bytes.
 */
void NXTBluetooth_readSync(U8* data, U32 size);
/** @} */

// Button functions
U8 NXTButton_read(void);

// Debug functions
#define DEBUG 0

#if DEBUG > 0
#define NXT_ASSERT(assertion, error_msg) \
    if (!(assertion)) { \
      NXT_printString(error_msg); \
      NXT_printString(" in "); \
      NXT_printString(__func__); \
      NXT_printString("(), "); \
      NXT_printString(__FILE__); \
      NXT_printString(" line "); \
      NXT_printNumber(__LINE__); \
      NXT_error("\n"); \
}
#else
#define NXT_ASSERT(assertion, error_msg)
#endif

#if DEBUG > 1
#define NXT_debugString(string) NXT_printString(string)
#define NXT_debugNumber(number) NXT_printNumber(number)
#define NXT_debugHexNumber(number) NXT_printHexNumber(number)
#else
#define NXT_debugString(string)
#define NXT_debugNumber(number)
#define NXT_debugHexNumber(number)
#endif

void NXT_init(void);
void NXT_error(const char* error_message);
void NXT_shutdown(void);
int NXT_rand(void);
U32 NXT_getSystemTime(void);
void NXTSystem_shutdown(void);
void NXTSystem_installWatchDog(NXTClosure watchdog);

U16 NXTSensor_read(U8 sensor);
void NXTSensor_enable(U8 sensor);
void NXTSensor_disable(U8 sensor);
void NXTRadar_enable(U8 sensor);
void NXTRadar_disable(U8 sensor);
U32 NXTRadar_readDistance(U8 sensor, U8 object);
bool NXTRadar_detect(U8 sensor);

S32 NXTMotor_getRotation(U8 sensor);
void NXTMotor_resetRotation(U8 sensor);
void NXTMotor_rotate(U8 motor, S8 percent);
void NXTMotor_rotateByAngle(U8 motor, S8 percent, U32 angle, bool hardBrake);
void NXTMotor_rotateForMilliseconds(U8 motor, S8 percent, U32 ms, bool hardBrake);
void NXTMotor_stop(U8 motor, bool hardBrake);

void NXTSpeaker_tone(U32 frequency, U32 ms);
void NXTSpeaker_asyncTone(U32 frequency, U32 ms);

#endif // _NXT_API_H
