/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <unistd.h>
#include <pthread.h>

#include <curses.h>
#include <time.h>

#include <sys/mman.h>
#define MAP_ANONYMOUS 0x20

#undef bool
#undef FALSE
#undef TRUE
#include "nxt_api.h"

#define NX_USERSPACE_SIZE 50112
#define NX_USERSPACE_END (NX_USERSPACE_START + NX_USERSPACE_SIZE)

/************************** Start of USB emulation *****************************/
#define NX_USB_PACKET_SIZE 64

int usb_socket = 0;
int server_socket = 0;
pthread_t usb_thread;

void* accept_usb_emulation_connections(void* arg) {
  struct sockaddr_in server_address, client_address;
  server_address.sin_family = AF_INET;
  server_address.sin_addr.s_addr = INADDR_ANY;
  server_address.sin_port = htons(10000);

  NXT_printString("Creating USB emulation socket\n");
  server_socket = socket(AF_INET, SOCK_STREAM, 0);
  if (server_socket < 0) {
    NXT_error("Error on opening socket for USB emulation\n");
  }

  NXT_printString("Binding USB emulation socket\n");
  if (bind(server_socket, (struct sockaddr*)&server_address, sizeof(server_address)) < 0) {
    NXT_error("Error on binding");
  }
  listen(server_socket, 5);

  NXT_printString("Waiting for socket connection for USB emulation on port 10000\n");

  while (TRUE) {
    int client_length = sizeof(client_address);
    int old_usb_socket = usb_socket;
    usb_socket = accept(server_socket, (struct sockaddr*)&client_address, &client_length);
    if (usb_socket < 0) {
      NXT_error("Error on accept");
    } else {
      // allow only one at a time
      if (old_usb_socket) {
        close(old_usb_socket);
      }
      NXT_printString("Got a new USB emulation connection\n");
    }
  }
}

void NXTUsb_init() {
  pthread_create(&usb_thread, NULL, accept_usb_emulation_connections, NULL);
}

void NXTUsb_shutdown() {
  if (server_socket) {
    close(server_socket);
  }
  if (usb_socket) {
    close(usb_socket);
  }
}

U8 NXTUsb_isConnected() {
  return usb_socket != 0;
}

U32 NXTUsb_packetSize() {
  return NX_USB_PACKET_SIZE;
}

int usb_bytes_read;

void NXTUsb_read(U8 *data, U32 length) {
  usb_bytes_read = recv(usb_socket, data, length, 0);
  if (usb_bytes_read < 0) {
    NXT_error("Error while reading from USB emulation socket!");
  }
}

U32 NXTUsb_dataRead() {
  return usb_bytes_read;
}

void NXTUsb_write(U8* data, U32 length) {
  send(usb_socket, data, length, 0);
}

U8 nx_usb_data_written() {
  return TRUE;
}

/************************** End of USB emulation *****************************/


/************************ Start of memory emulation **************************/

U8* NX_USERSPACE_START;

void NXTMemory_init() {
  NX_USERSPACE_START = (U8*)malloc(NX_USERSPACE_SIZE);
  // initialize with garbage (on the NXT this is sometimes the situation - perhaps after flashing and without turnoff
  U8* memory = NX_USERSPACE_START;
  while (memory < NX_USERSPACE_START + NX_USERSPACE_SIZE) {
    *memory++ = rand() % 256;
  }
}

void NXTMemory_shutdown() {
  free(NX_USERSPACE_START);
}

U8* NXTMemory_getStart() {
  return NX_USERSPACE_START;
}

U8* NXTMemory_getEnd() {
  return NX_USERSPACE_START + NX_USERSPACE_SIZE;
}

void NXTMemory_copy(void* dest, void* src, U32 length) {
  memcpy(dest, src, length);
}

void NXTMemory_move(void* dest, void* src, U32 length) {
  memmove(dest, src, length);
}

bool NXTMemory_isRAMAddress(void* address) {
  return address >= (void*)NXTMemory_getStart() &&
      address < (void*)NXTMemory_getEnd();
}

/************************** End of memory emulation **************************/


/************************* Start of Flash emulation **************************/
#define EFC_PAGES 1024
#define EFC_PAGE_WORDS 64
#define EFC_PAGE_BYTES 256
#define NX_FLASH_SIZE 256 * 1024

U8* NX_ROM_USERSPACE_START;
U8* NX_ROM_USERSPACE_END;
FILE* flash_file;

void NXTFlash_init() {
  NXT_printString("Initializing Flash ROM emulation\n");
  NX_ROM_USERSPACE_START = (U8*)mmap(NULL, NX_FLASH_SIZE, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
  NX_ROM_USERSPACE_END = NX_ROM_USERSPACE_START + NX_FLASH_SIZE;
  flash_file = fopen("flash_emulation.dat", "rb+");
  if (flash_file == NULL) {
    flash_file = fopen("flash_emulation.dat", "wb+");
    if (flash_file == NULL) {
      NXT_error("Could not create flash emulation file!");
    }
    for (int i = 0; i < EFC_PAGES * EFC_PAGE_BYTES; i++) {
      fputc(0, flash_file);
    }
  } else {
    if (!fread(NX_ROM_USERSPACE_START, 1, NX_FLASH_SIZE, flash_file)) {
      NXT_error("File read error");
    }
  }
  mprotect(NX_ROM_USERSPACE_START, NX_FLASH_SIZE, PROT_READ);
}

void NXTFlash_shutdown() {
  if (flash_file)
    fclose(flash_file);
}

void NXTFlash_readPage(U32 page, U8* data) {
  memcpy(data, NX_ROM_USERSPACE_START + page * EFC_PAGE_BYTES, EFC_PAGE_BYTES);
}

void NXTFlash_writePage(U8* data, U32 page) {
  mprotect(NX_ROM_USERSPACE_START, NX_FLASH_SIZE, PROT_READ | PROT_WRITE);
  memcpy(NX_ROM_USERSPACE_START + page * EFC_PAGE_BYTES, data, EFC_PAGE_BYTES);
  fseek(flash_file, page * EFC_PAGE_BYTES, SEEK_SET);
  if (!fwrite(data, 1, EFC_PAGE_BYTES, flash_file)) {
    NXT_error("File write error");
  }
  fflush(flash_file);
  mprotect(NX_ROM_USERSPACE_START, NX_FLASH_SIZE, PROT_READ);
}

U32 NXTFlash_getPageSize() {
  return EFC_PAGE_BYTES;
}

U32 NXTFlash_firstPage() {
  return 0;
}

U32 NXTFlash_lastPage() {
  return EFC_PAGES - 1;
}

U8* NXTFlash_getStart() {
  return NX_ROM_USERSPACE_START;
}

U8* NXTFlash_getEnd() {
  return NX_ROM_USERSPACE_START + NX_FLASH_SIZE;
}

/************************* End of Flash emulation ****************************/

/************************* Start of display emulation ************************/

pthread_mutex_t display_mutex;

void NXTDisplay_init() {
  initscr();
  nodelay(stdscr, TRUE);
  keypad(stdscr, TRUE);
  scrollok(stdscr, TRUE);
  noecho();
  curs_set(0);
  pthread_mutex_init(&display_mutex, NULL);
}

void NXTDisplay_shutdown() {
  endwin();
}

void NXTDisplay_setScrollMode(bool on) {
}

void NXTDisplay_clear() {
  clear();
  refresh();
}

U8 NXTDisplay_getCursorX() {
  int x;
  int y;
  getyx(stdscr, y, x);
  return x;
}

U8 NXTDisplay_getCursorY() {
  int x;
  int y;
  getyx(stdscr, y, x);
  return y;
}

void NXTDisplay_setCursor(U8 x, U8 y) {
  move(y, x);
  refresh();
}

void NXT_printChar(char character) {
  pthread_mutex_lock(&display_mutex);
  printw("%c", character);
  refresh();
  pthread_mutex_unlock(&display_mutex);
}

void NXT_printString(const char* string) {
  pthread_mutex_lock(&display_mutex);
  printw("%s", string);
  refresh();
  pthread_mutex_unlock(&display_mutex);
}

void NXT_printNumber(U32 i) {
  pthread_mutex_lock(&display_mutex);
  printw("%i", i);
  refresh();
  pthread_mutex_unlock(&display_mutex);
}

void NXT_printHexNumber(U32 i) {
  pthread_mutex_lock(&display_mutex);
  printw("%X", i);
  refresh();
  pthread_mutex_unlock(&display_mutex);
}

/************************** End of display emulation *************************/

/**************************** Sensor emulation *******************************/

void NXTSensor_enable(U8 sensor) {
}

void NXTSensor_disable(U8 sensor) {
}

U16 NXTSensor_read(U8 sensor) {
  return 600;
}

bool NXTRadar_detect(U8 sensor) {
  return TRUE;
}

void NXTRadar_enable(U8 sensor) {
}

void NXTRadar_disable(U8 sensor) {
}

U32 NXTRadar_readDistance(U8 sensor, U8 object) {
  return 255;
}

/************************** End of sensor emulation **************************/

/****************************** Motor emulation ******************************/

void NXTMotor_rotate(U8 motor, S8 percent) {
}

void NXTMotor_stop(U8 motor, bool hardBrake) {
}

void NXTMotor_rotateForMilliseconds(U8 motor, S8 percent, U32 milliseconds, bool hardBrake) {
}

void NXTMotor_rotateByAngle(U8 motor, S8 percent, U32 angle, bool hardBrake) {
}

S32 NXTMotor_getRotation(U8 motor) {
  return 0;
}

void NXTMotor_resetRotation(U8 motor) {
}

/*************************** End of motor emulation **************************/

/************************** Bluetooth emulation ******************************/

void NXTBluetooth_init() {
}

void NXTBluetooth_shutdown() {
}

void NXTBluetooth_setDiscoverable(bool on) {
}

bool NXTBluetooth_isConnectionPending() {
}

bool NXTBluetooth_isWaitingForPIN() {
}

void NXTBluetooth_sendPin(char* pin) {
}

void NXTBluetooth_acceptConnection(bool accept) {
}

S32 NXTBluetooth_connectionEstablished() {
}

void NXTBluetooth_openStream(S32 handle) {
}

void NXTBluetooth_closeStream() {
}

bool NXTBluetooth_isStreamOpen() {
}

void NXTBluetooth_read(U8* data, U32 length) {
}

U32 NXTBluetooth_dataRead() {
}

void NXTBluetooth_write(U8* data, U32 length) {
}

bool NXTBluetooth_dataWritten() {
}

/************************ End of Bluetooth emulation *************************/

U8 NXTButton_read() {
  U8 buttons;
  static last_key = BUTTON_NONE;
  static last_key_start = 0;
  int key = getch();
  switch (key) {
    case ERR: {
      if (NXT_getSystemTime() - last_key_start > 50) {
        last_key = BUTTON_NONE;
      }
      break;
    }
    case KEY_LEFT: last_key = BUTTON_LEFT; break;
    case KEY_RIGHT: last_key = BUTTON_RIGHT; break;
    case 10: last_key = BUTTON_OK; break;
    case 'q': last_key = BUTTON_CANCEL; break;
  }
  if (last_key != BUTTON_NONE && key != ERR) {
    last_key_start = NXT_getSystemTime();
  }
  return last_key;
}

static inline long get_time() {
  struct timeval time;
  gettimeofday(&time, NULL);
  return time.tv_usec / 1000 + time.tv_sec * 1000;
}

U32 NXT_getSystemTime() {
  static long startTime = 0;
  if (startTime == 0) {
    startTime = get_time();
  }
  return get_time() - startTime;
}

void NXTSpeaker_tone(U32 frequency, U32 ms) {
}

void NXTSpeaker_asyncTone(U32 frequency, U32 ms) {
}

void NXTSystem_installWatchDog(NXTClosure watchdog) {
}

void NXTSystem_shutdown() {
  exit(-1);
}
