/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

/** \mainpage
 *
 * This is the generated documentation for the NXTalk virtual machine. It
 * describes the modules, data structures, functions and files defined and
 * used within the NXTalk sources.
 */

#include "object.h"
#include "image.h"
#include "interpreter.h"
#include "method.h"

static void loadImage() {
  NXT_printString("Waiting for new image...\n");
  while (!NXTUsb_isConnected());
  U32 byteCount = 0;
  NXTUsb_readSync((U8*)&byteCount, 4);

  NXTDisplay_clear();
  NXT_printString("Receiving image... ");
  U8 cursorX = NXTDisplay_getCursorX();
  U8 cursorY = NXTDisplay_getCursorY();
  NXT_printString("0%");

  U8 page[NXTFlash_getPageSize()];
  U32 bytesRead = 0;
  U32 pageNumber = NXTFlash_firstPage();
  do {
    // Read exactly one flash page from USB port
    U32 thisPage = 0;
    do {
      NXTUsb_read(page + thisPage, MIN(MIN(NXTUsb_packetSize(), NXTFlash_getPageSize() - thisPage), byteCount - bytesRead));
      while (!NXTUsb_dataRead());
      thisPage += NXTUsb_dataRead();
      bytesRead += NXTUsb_dataRead();
    } while (thisPage < NXTFlash_getPageSize() && bytesRead < byteCount);
    NXTFlash_writePage(page, pageNumber++);
    NXTDisplay_setCursor(cursorX, cursorY);
    NXT_printNumber(bytesRead * 100 / byteCount);
    NXT_printString("%");
  } while (bytesRead < byteCount);
  NXTDisplay_setCursor(cursorX, cursorY);
  NXT_printString("Done.\n");
  NXT_printString("Receiving object table... ");
  cursorX = NXTDisplay_getCursorX();
  cursorY = NXTDisplay_getCursorY();
  NXT_printString("0%");
  U32 numberOfObjects;
  NXTUsb_readSync((U8*)&numberOfObjects, 4);
  byteCount = numberOfObjects;
  bytesRead = 0;
  pageNumber = NXTFlash_lastPage();
  do {
    U32 thisPage = 0;
    do {
      NXTUsb_read(page + thisPage, MIN(MIN(NXTUsb_packetSize(), NXTFlash_getPageSize() - thisPage), byteCount - bytesRead));
      while (!NXTUsb_dataRead());
      thisPage += NXTUsb_dataRead();
      bytesRead += NXTUsb_dataRead();
    } while (thisPage < NXTFlash_getPageSize() && bytesRead < byteCount);
    // swap order of object table
    U32* words = (U32*)page;
    for (U32 i = 0; i < NXTFlash_getPageSize() / 4 / 2; i++) {
      U32 temp = words[i];
      words[i] = words[NXTFlash_getPageSize() / 4 - 1 - i];
      words[NXTFlash_getPageSize() / 4 - 1 - i] = temp;
    }
    // write page to flash
    NXTFlash_writePage(page, pageNumber--);
    NXTDisplay_setCursor(cursorX, cursorY);
    NXT_printNumber(bytesRead * 100 / byteCount);
    NXT_printString("%");
  } while (bytesRead < byteCount);
  NXTDisplay_setCursor(cursorX, cursorY);
  NXT_printString("Done.\n");
}

static int startOfPress = 0;

static void watchdog(void) {
  if (NXTButton_read() == BUTTON_CANCEL) {
    if (startOfPress == 0) {
      startOfPress = NXT_getSystemTime();
    }
    if (NXT_getSystemTime() - startOfPress >= 3000) {
      NXT_shutdown();
    }
  } else {
    startOfPress = 0;
  }
}

void main(void) {
  NXT_init();
  NXTSystem_installWatchDog(watchdog);

  NXT_printString("NXTalk 0.7\n-----------\n");
  NXTButton lastButton = BUTTON_NONE;

  NXT_printString("Ok: Boot\nRight: New image\nCancel: Exit\n");
  //nx_display_uint(NX_USERSPACE_SIZE);
  while(TRUE)
  {
    lastButton = NXTButton_read();
    if (lastButton == BUTTON_OK || lastButton == BUTTON_RIGHT) {
      break;
    }
    if (lastButton == BUTTON_CANCEL) {
      NXT_debugString("Shutting down...");
      NXT_shutdown();
    }
  }

  // wait till button is released
  while (NXTButton_read() != BUTTON_NONE) {
  }
  if (lastButton == BUTTON_RIGHT) {
    loadImage();
  }
  // If no image is found, wait for USB transfer
  if (!Image_isPresent()) {
    NXT_printString("No image found! ");
    loadImage();
  }

  Image_initializeObjectMemory();
  Interpreter_initialize();
  NXT_printString("Booting image...\n");

  /*NXT_printNumber(objectCount);
  while (NXTButton_read() == BUTTON_NONE);*/

  Interpreter_stackPush(specialObjects->Processor);
  Interpreter_send(specialObjects->Processor,
                    specialObjects->startupSymbol, FALSE);
  // This call never returns
  Interpreter_run();
}
