/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

#include "object.h"

#include "image.h"
#include "integer.h"
#include "garbagecollection.h"
#include "interpreter.h"
#include "primitives.h"
#include "freelist.h"

Object* Object_getForWrite(ObjectRef objectRef) {
  if (ObjectTable_isROMObject(objectRef)) {
    Image_moveObjectToRAM(objectRef);
  }
  return ObjectTable_getAddress(objectRef);
}

ObjectRef Object_getClass(ObjectRef object) {
  if (!ObjectTable_isRealObject(object)) {
    return specialObjects->NXTSmallInteger;
  }
  return Object_get(object)->klass;
}

U16 Object_getSize(ObjectRef object) {
  Class* class = Class_get(Object_getClass(object));
  if (Object_isIndexable(object)) {
    return ((IndexableObject*)ObjectTable_getAddress(object))->size;
  } else {
    return Class_getInstVarCount(class);
  }
}

U16 Object_getElementSize(ObjectRef object) {
  return Class_getElementSize(Class_get(Object_getClass(object)));
}

U16 Object_getFullSize(ObjectRef object) {
  U16 size = Object_getSize(object) * Object_getElementSize(object) + 2;
  if (Class_isVariable(Class_get(Object_getClass(object)))) {
    size += 2;
  }
  size += size & 0x1;
  return size;
}

ObjectRef Object_getInstVar(ObjectRef object, int index) {
  Object* objectPtr = ObjectTable_getAddress(object);
  if (Object_getElementSize(object) != 2 &&
      objectPtr->klass != specialObjects->NXTBlock &&
      objectPtr->klass != specialObjects->NXTMethod) {
    if (Object_getElementSize(object) == 1) {
      return SmallInteger_toReference(ByteObject_get(object)->bytes[index]);
    }
    NXT_error("Trying to load inst var from an object containing no object references");
  }
  if (index < 0 || index >= Object_getSize(object)) {
    NXT_error("Access out of bounds");
  }
  if (Object_isIndexable(object)) {
    return ((IndexableObject*)objectPtr)->instanceVariables[index];
  } else {
    return objectPtr->instanceVariables[index];
  }
}

static ObjectRef Object_privSetInstVar(ObjectRef object, int index, ObjectRef value, bool doRefCounting) {
  if (ObjectTable_isROMObject(object)) {
    Image_moveObjectToRAM(object);
  }
  Object* objectPtr = Object_get(object);
  if (Object_getElementSize(object) == 1) {
    ((ByteObject*)objectPtr)->bytes[index] = SmallInteger_fromReference(value);
    return value;
  }
  ObjectRef* instVar = Object_isIndexable(object) ? &((IndexableObject*)objectPtr)->instanceVariables[index] : &objectPtr->instanceVariables[index];
  if (doRefCounting) {
    Object_incReferenceCount(value);
    Object_decReferenceCount(*instVar);
  }
  *instVar = value;
  return value;
}

ObjectRef Object_setInstVar(ObjectRef object, int index, ObjectRef value) {
  return Object_privSetInstVar(object, index, value, TRUE);
}

void Object_free(ObjectRef object) {
  ClassRef class = Object_getClass(object);
  if (class == specialObjects->NXTBlockFrame ||
      class == specialObjects->NXTMethodFrame) {
    // Frames need special handling, because they come from a global pool
    Frame_free(object);
  } else {
    Object_forceFree(object);
  }
}

void Object_forceFree(ObjectRef object) {
  NXT_ASSERT(ObjectTable_isRAMObject(object), "Freeing of non-RAM object");
  if (ObjectTable_isReferencingInstances(object)) {
          // decrement reference count of depending objects
    Object* self = ObjectTable_getAddress(object);
          //Object_decReferenceCount(self->klass);
    NXT_ASSERT(self, "Null pointer");

    int size;
    ObjectRef* instanceVariable;
    if (Object_isIndexable(object)) {
      size = ((IndexableObject*)self)->size;
      instanceVariable = ((IndexableObject*)self)->instanceVariables;
    } else {
      size = Class_getInstVarCount(Class_get(self->klass));
      instanceVariable = self->instanceVariables;
    }

    while (size--) {
      ObjectRef reference = *instanceVariable++;
      Object_decReferenceCount(reference);
    }
  }
  NXT_ASSERT(!ObjectTable_isFree(object), "Double free");
  NXT_ASSERT(!ObjectTable_isRenewed(object), "Freeing a reference moved from ROM");
  Freelist_free(object, Object_getFullSize(object));
  ObjectTable_setFree(object, TRUE);
  objectCount--;
}

void Object_decReferenceCount(ObjectRef object) {
  // TODO: Make this not recursive! (see bluebook)
  if (ObjectTable_isRealObject(object) && object != specialObjects->nil) {
    int count = ObjectTable_getReferenceCount(object);
    NXT_ASSERT(count > 0, "Ref count decrement on object without refs");
    if (count < RC_TOO_MANY_REFERENCES) {
      count--;
      if (count == 0) {
        // free the object
        Object_free(object);
      }
      ObjectTable_setReferenceCount(object, count);
    }
  }
}

bool Object_isClass(ObjectRef object) {
  if (object == specialObjects->NXTMetaclass) return TRUE;
  ObjectRef klass = Object_getClass(object);
  if (klass == specialObjects->NXTClass) return TRUE;
  if (klass == specialObjects->NXTMetaclass) return TRUE;
  if (Object_getClass(klass) == specialObjects->NXTMetaclass) return TRUE;
  return FALSE;
}

void Object_setClass(ObjectRef object, ObjectRef value) {
  Object* self = ObjectTable_getAddress(object);
  Object_incReferenceCount(value);
  Object_decReferenceCount(self->klass);
  self->klass = value;
}

ObjectRef ObjectMemory_findNextInstanceFor(ObjectRef object) {
  ObjectRef class = Object_getClass(object);
  while (object < MAX_REFERENCE) {
    object += 2;
    if (!ObjectTable_isFree(object) && Object_get(object)->klass == class) {
      return object;
    }
  }
  return specialObjects->nil;
}

void Object_primHash(PrimitiveFrame* frame) {
  Primitive_return(frame, Integer_new((frame->self >> 1) % (SMALLINTEGER_MAX + 1)));
}

void Object_primClass(PrimitiveFrame* frame) {
  ObjectRef result = Object_getClass(frame->self);
  Primitive_return(frame, result);
}

void Object_primSetClass(PrimitiveFrame* frame) {
  Object_setClass(frame->self, Primitive_argumentAt(frame, 0));
  Primitive_return(frame, frame->self);
}

void Object_primSize(PrimitiveFrame* frame) {
  ObjectRef result = SmallInteger_toReference(Object_getSize(frame->self));
  Primitive_return(frame, result);
}

void Object_primAt(PrimitiveFrame* frame) {
  U16 index = SmallInteger_fromReference(Primitive_argumentAt(frame, 0));
  ObjectRef result = Object_getInstVar(frame->self, index - 1);
  Primitive_return(frame, result);
}

void Object_primAtPut(PrimitiveFrame* frame) {
  ObjectRef object = Primitive_argumentAt(frame, 1);
  U16 index = SmallInteger_fromReference(Primitive_argumentAt(frame, 0));
  Primitive_return(frame, Object_setInstVar(frame->self, index - 1, object));
}

void Object_primEquals(PrimitiveFrame* frame) {
  ObjectRef other = Primitive_argumentAt(frame, 0);
  if (frame->self == other) {
    Primitive_return(frame, specialObjects->true);
  } else {
    Primitive_return(frame, specialObjects->false);
  }
}

void Object_primPerform(PrimitiveFrame* frame) {
  int argumentCount = frame->argumentCount - 1;
  ObjectRef symbolToPerform = Primitive_argumentAt(frame, 0);
  if (argumentCount != Symbol_countArguments(symbolToPerform)) {
    NXT_error("Wrong argument count for #perform:!");
  }
  if (argumentCount > 0) {
    // Move every argument on the stack one step down
    for (int i = 0; i < argumentCount; i++) {
      Primitive_argumentAtPut(frame, i, Primitive_argumentAt(frame, i + 1));
    }
    Interpreter_stackPop();
  }
  Interpreter_send(frame->self, symbolToPerform, FALSE);
  Object_decReferenceCount(symbolToPerform);
}

void Object_primPerformWithArguments(PrimitiveFrame* frame) {
  ObjectRef arguments = Interpreter_stackPop(); // Primitive_argumentAt(frame, 1);
  int argumentCount = Object_getSize(arguments);
  ObjectRef symbolToPerform = Interpreter_stackPop(); // Primitive_argumentAt(frame, 0);
  if (argumentCount != Symbol_countArguments(symbolToPerform)) {
    NXT_error("Wrong argument count for #perform:withArguments:!");
  }
  for (int i = 0; i < argumentCount; i++) {
    ObjectRef argument = Object_getInstVar(arguments, i);
    Object_incReferenceCount(argument);
    Interpreter_stackPush(argument);
  }
  Interpreter_send(frame->self, symbolToPerform, FALSE);
  Object_decReferenceCount(arguments);
  Object_decReferenceCount(symbolToPerform);
}

void Object_primHalt(PrimitiveFrame* frame) {
  // only for VM debugging purposes
  Primitive_return(frame, frame->self);
}

void Object_primNextInstance(PrimitiveFrame* frame) {
  Primitive_return(frame, ObjectMemory_findNextInstanceFor(frame->self));
}
