/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

/** \file method.h
 *  \brief Functions for objects.
 */

#ifndef _OBJECT_H_
#define _OBJECT_H_

#include "nxt_api.h"
#include "objecttable.h"

/** \brief Type redefinition to indicate Object references. */
typedef U16 ObjectRef;

/** \brief Regular object representation in memory.
 *
 * A typical NXTalk object consists of a class reference and a number of
 * instance variables specified by the class.
 */
struct Object {

  /** \brief Class reference. */
  ObjectRef klass;

  /** \brief Instance variables. */
  ObjectRef instanceVariables[];
};
typedef struct Object Object;

/** \brief Indexable object representation in memory.
 *
 * NXTalk objects may be indexable, which means that their data is accessed
 * by index instead of named variables. As different instances of the same
 * class can thus have different sizes, they additionally carry the size
 * information.
 */
struct IndexableObject {

  /** \brief Class reference. */
  ObjectRef klass;

  /** \brief Size of this object.
   *
   * The size is the number of following references. Hence, to calculate
   * the full size of an object, the value has to be multiplied with 2 and
   * the 4 leading bytes have to be added.
   */
  U16 size;

  /** \brief Instance variables. */
  ObjectRef instanceVariables[];
};
typedef struct IndexableObject IndexableObject;

/** \brief Byte-oriented object representation in memory.
 *
 * NXTalk objects may contain only bytes instead of references. As this
 * automatically implicates some kind of byte array, these objects are
 * always indexable per definition.
 */
struct ByteObject {

  /** \brief Class reference. */
  ObjectRef klass;

  /** \brief Size of this object.
   *
   * The number of bytes.
   */
  U16 size;

  /** \brief The bytes. */
  U8 bytes[];
};
typedef struct ByteObject ByteObject;

/** \brief Resolve an Object reference.
 *
 * Retrieves a direct pointer to an object.
 *
 * \param objectRef The reference to the object.
 * \return The direct pointer to the object in memory.
 */
static inline Object* Object_get(ObjectRef objectRef) {
  return (Object*)ObjectTable_getAddress(objectRef);
}

/** \brief Resolve an Object reference for writing.
 *
 * Retrieves a direct pointer to an object ensuring it lies in the
 * writeable RAM section.
 *
 * \param reference The reference to the object.
 * \return The direct pointer to the object in RAM.
 */
Object* Object_getForWrite(ObjectRef reference);

/** \brief Resolve an IndexableObject reference.
 *
 * Retrieves a direct pointer to an indexable object.
 *
 * \param objectRef The reference to the indexable object.
 * \return The direct pointer to the indexable object in memory.
 */
static inline IndexableObject* IndexableObject_get(ObjectRef objectRef) {
  return (IndexableObject*)ObjectTable_getAddress(objectRef);
}

/** \brief Resolve a ByteObject reference.
 *
 * Retrieves a direct pointer to a byte object.
 *
 * \param objectRef The reference to the byte object.
 * \return The direct pointer to the byte object in memory.
 */
static inline ByteObject* ByteObject_get(ObjectRef objectRef) {
  return (ByteObject*)ObjectTable_getAddress(objectRef);
}

/** \brief Frees an object.
 *
 * Objects not needed anymore (e.g. because they are not referenced) can
 * be freed. This means, their memory is available for reuse on later
 * object allocations.
 *
 * \param object The object, which shall be deallocated.
 */
void Object_free(ObjectRef object);

/** \brief Definately frees an object.
 *
 * Object_free() only frees an object, if it is not a Frame context. In that
 * case, it tries to relinquish it back to the frame pool. This function
 * omits this check and always deletes the object.
 *
 * \param object The object to be deleted.
 */
void Object_forceFree(ObjectRef object);

/** \brief Increments the reference count.
 *
 * Increments the reference count of the given object. If it reaches 31,
 * the reference count stays the same because there are only 5 bits in
 * the object table for the RC.
 *
 * \param object The object.
 */
static inline void Object_incReferenceCount(ObjectRef object) {
  if (ObjectTable_isRealObject(object)) {
    ObjectTable_incReferenceCount(object);
  }
}

/** \brief Decrement the reference count.
 *
 * \param object The object.
 */
void Object_decReferenceCount(ObjectRef object);

/** \brief Finds the next instance of this object's class.
 *
 * In fact, this method finds the object with next ascending
 * reference number, which has the same class.
 *
 * \param object This object.
 * \return The next object in memory with the same class.
 */
ObjectRef ObjectMemory_findNextInstanceFor(ObjectRef object);

/** \name Getters
 * @{
 */

/** \brief Returns the size.
 *
 * The size of an object is defined as the number of its accessable data
 * slots. Hence, while for a string it is the number of (byte) characters,
 * for a regular instance it is the number of instance variables.
 *
 * \param object The object.
 * \return Size of the object.
 */
U16 Object_getSize(ObjectRef object);

/** \brief Returns the element size.
 *
 * Returns the element size of one data slot in the object. This is one for
 * byte-oriented objects and two for regular objects consisting of object
 * references.
 */
U16 Object_getElementSize(ObjectRef object);

/** \brief Returns the full size.
 *
 * Returns the full size of the object in bytes including a padding byte
 * if necessary.
 *
 * \param object The object.
 * \return The full size in bytes.
 */
U16 Object_getFullSize(ObjectRef object);

/** \brief Returns the contents of an instance variable.
 *
 * Returns the contents of an instance variable, regardless whether the object
 * is indexable or byte-oriented.
 *
 * \param object The object.
 * \param number Index of the instance variable or data slot.
 * \return  The reference stored in the instance variable or a tagged
 *          integer representing the byte value.
 */
ObjectRef Object_getInstVar(ObjectRef object, int number);

/** \brief Returns the class of an object.
 *
 * Returns the class of an object, also recognizes tagged-integers.
 *
 * \param object The object reference.
 * \return A reference to the object's class.
 */
ObjectRef Object_getClass(ObjectRef object);

/** \brief Returns, whether the object is indexable or not.
 *
 * \param object The object reference.
 * \return TRUE, if the object is indexable, FALSE if not.
 */
static inline bool Object_isIndexable(ObjectRef object) {
  return ObjectTable_isIndexable(object);
}

/** \brief Returns, whether the object is a class or not.
 *
 * \param object The object reference.
 * \return TRUE, if the object is a class, FALSE if not.
 */
bool Object_isClass(ObjectRef object);
/** @} */

/** \name Setters
 * @{
 */

/** \brief Sets the contents of an instance variable.
 *
 * Sets the contents of an instance variable, regardless whether the object
 * is indexable or byte-oriented. It also does necessary reference counting.
 *
 * \param object  The object.
 * \param index   Index of the instance variable or data slot.
 * \param value   The new value.
 */
ObjectRef Object_setInstVar(ObjectRef object, int index, ObjectRef value);

/** \brief Sets the class.
 *
 * \param object  The object.
 * \param value   A reference to the new class.
 */
void Object_setClass(ObjectRef object, ObjectRef value);
/** @} */

/** \name Primitives
 * @{
 */

/** Forward declaration of PrimitiveFrame */
typedef struct PrimitiveFrame PrimitiveFrame;

/** \brief Calculates the hash of the object. */
void Object_primHash(PrimitiveFrame* frame);

/** \brief Returns the class of an object. */
void Object_primClass(PrimitiveFrame* frame);

/** \brief Sets the class of an object. */
void Object_primSetClass(PrimitiveFrame* frame);

/** \brief Returns the size of an object. */
void Object_primSize(PrimitiveFrame* frame);

/** \brief Returns an instance variable or slot of an object. */
void Object_primAt(PrimitiveFrame* frame);

/** \brief Updates an instance variable or slot of an object. */
void Object_primAtPut(PrimitiveFrame* frame);

/** \brief Tests for equality to another object. */
void Object_primEquals(PrimitiveFrame* frame);

/** \brief Sends a message to the object. */
void Object_primPerform(PrimitiveFrame* frame);

/** \brief Sends a message to the object and takes the arguments as array. */
void Object_primPerformWithArguments(PrimitiveFrame* frame);

/** \brief For VM debugging purposes. */
void Object_primHalt(PrimitiveFrame* frame);

/** \brief Returns the next instance of the same class in memory. */
void Object_primNextInstance(PrimitiveFrame* frame);
/** @} */

#endif //_OBJECT_H_
