/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

/** \file objecttable.h
 * \brief Functions and variables for the object table.
 * \author Martin Beck
 *
 * The NXTalk VM uses an object table in order to save space in the actual
 * object data. Object references are indices to the table and consume
 * two bytes. An actual direct pointer would take up four bytes. As the
 * NXT has very few memory, only the lower 22 bits of an address are
 * really used leaving the rest for payload.
 *
 * An object table entry thus consists of only four bytes, which hold the
 * address of the referred object. Furthermore, they also contain the
 * reference count and a marked bit for the object used by the garbage.
 * collection. Another bit indicates free memory regions not used by any
 * object. Next, a bit marks objects, which have been moved from the
 * Flash memory to the RAM. Finally, the last two bits are used to speed
 * up object processing by containing two important hints about the
 * object structure, which would have to be fetched from the object's
 * class otherwise.
 *
 * The object table itself is place at the top of the RAM. Thus, it is
 * growing downwards, if needed and reaches the object heap at some point
 * in time. When this happens, garbage collection should be invoked.
 * Furthermore, object references are handed out starting at 65535. Thus,
 * no real subtraction is needed for the reversed table structure. However,
 * object references are always odd numbers as even numbers are used as
 * tagged integers. Thus, references are shifted to the right by one, before
 * they are used as index into the object table.
 */

#ifndef _OBJECTTABLE_H_
#define _OBJECTTABLE_H_

#include "nxt_api.h"

/** \brief Number of bits in the object table used for reference counting. */
#define REFERENCE_COUNT_BITS 5

/** \brief Index of the object table bit used for indicating free addresses. */
#define FREE_BIT (32 - REFERENCE_COUNT_BITS - 1)

/** \brief Index of the object table bit used for indicating marked objects. */
#define MARK_BIT (FREE_BIT - 1)

/** \brief Index of the object table bit used for renewed objects.
 *
 * This bit indicates objects, which have been moved from the Flash to the
 * writeable RAM region. This information is required by the mark & sweep
 * garbage collection as it only collects RAM objects. As objects coming
 * from the Flash may not be referenced by any other RAM object, they
 * have to be identified explicitly.
 */
#define RENEWED_BIT (MARK_BIT - 1)

/** \brief Index of the object table bit used to indicate indexable objects. */
#define INDEXABLE_BIT (RENEWED_BIT - 1)

/** \brief Index of the object table bit used to indicate regular objects.
 *
 * This bit indicates objects, that have regular instance variables
 * referencing other objects. This avoids a class lookup of this information
 * in some cases like object traversion.
 */
#define INSTANCES_BIT (INDEXABLE_BIT - 1)

/** \brief Index of the bit in the object table indicating RAM addresses. */
#define RAM_BIT (INSTANCES_BIT - 1)

/** \brief Helper macro for modifying exactly one bit.
 *
 * \param oldValue  The value to be modified.
 * \param bitNumber Index of the bit to be modified.
 * \param newBit    New value for the bit.
 * \return The modified value.
 */
#define MODIFY_BIT(oldValue, bitNumber, newBit) \
  ((oldValue & ~(1 << bitNumber)) | (newBit << bitNumber))

/** \brief Value indicating too many references to continue counting. */
#define RC_TOO_MANY_REFERENCES ((1 << REFERENCE_COUNT_BITS) - 1)

/** \brief Amount of bits to shift object table value right for the RC. */
#define RC_SHIFTER (32 - REFERENCE_COUNT_BITS)

/** \brief Bit mask to filter out only the reference count bits. */
#define RC_BIT_MASK (RC_TOO_MANY_REFERENCES << RC_SHIFTER)

/** \brief Maximum value for an object reference. */
#define MAX_REFERENCE 65535

/** \brief The object table. */
extern U32* objectTable;

/** \brief The start of the object table in RAM. */
extern U32* objectTableStart;

/** \brief The start of the object table in ROM. */
extern U32* objectTableROMStart;

/** \brief Lowest reference used by the object table. */
extern U16 minReference;

/** \brief Count of objects currently in use. */
extern U16 objectCount;

/** \brief Checks whether a reference points to a real object.
 *
 * \param reference The reference.
 * \return  TRUE, if the reference points to a real object. FALSE,
 *          if it indicates a tagged-integer.
 */
static inline bool ObjectTable_isRealObject(U16 reference) {
  return reference & 0x1;
}

/** \brief Fetches a value from the object table.
 *
 * Returns the object table entry corresponding to the given
 * object reference. In assertion mode, it also checks,
 * whether the given reference is valid.
 *
 * \param reference An object reference.
 * \return The 32 bits stored in the object table.
 */
static inline U32 ObjectTable_get(U16 reference) {
  NXT_ASSERT(ObjectTable_isRealObject(reference), "no object reference");
  NXT_ASSERT(reference >= minReference, "wrong object reference");
  return objectTable[(U16)reference >> 1];
}

/** \brief Updates a value in the object table.
 *
 * Updates the object table entry corresponding to the given
 * object reference. In assertion mode it also checks, whether
 * the given reference is valid.
 *
 * \param reference An object reference.
 * \param value     The new 32-bit value.
 */
static inline void ObjectTable_set(U16 reference, U32 value) {
  NXT_ASSERT(ObjectTable_isRealObject(reference), "no object reference");
  NXT_ASSERT(reference >= minReference, "wrong object reference");
  objectTable[(U16)reference >> 1] = value;
}

/** \brief Returns only the reference count for a reference.
 *
 * \param reference An object reference.
 * \return The reference count.
 */
static inline int ObjectTable_getReferenceCount(U16 reference) {
  return ObjectTable_get(reference) >> RC_SHIFTER;
}

/** \brief Sets the reference count for a reference.
 *
 * \param reference An object reference.
 * \param count     The new reference count.
 */
static inline void ObjectTable_setReferenceCount(U16 reference, int count) {
  ObjectTable_set(reference, ((count & RC_TOO_MANY_REFERENCES) << RC_SHIFTER) |
      (ObjectTable_get(reference) & ~RC_BIT_MASK));
}

/** \brief Increments the reference count for a reference.
 *
 * \param reference An object reference.
 */
static inline void ObjectTable_incReferenceCount(U16 reference) {
  U32 value = ObjectTable_get(reference);
  value += 1 << RC_SHIFTER;
  if (value >= 1 << RC_SHIFTER) {
    ObjectTable_set(reference, value);
  }
}

/** \brief Indicates, whether a reference points to free memory.
 *
 * \param reference An object reference.
 * \return TRUE, if the reference points to free memory. FALSE, if not.
 */
static inline bool ObjectTable_isFree(U16 reference) {
  return ObjectTable_get(reference) == 0 || (ObjectTable_get(reference) >> FREE_BIT) & 1;
}

/** \brief Sets the free bit for a reference.
 *
 * \param reference An object reference.
 * \param isFree    The new value for the free bit.
 */
static inline void ObjectTable_setFree(U16 reference, bool isFree) {
  ObjectTable_set(reference, MODIFY_BIT(ObjectTable_get(reference), FREE_BIT, isFree));
}

/** \brief Indicates, whether an object is marked by the GC.
 *
 * \param reference An object reference.
 * \return TRUE, if the referenced object is marked. FALSE, if not.
 */
static inline bool ObjectTable_isMarked(U16 reference) {
  if (!ObjectTable_isRealObject(reference)) {
    return TRUE;
  }
  return (ObjectTable_get(reference) >> MARK_BIT) & 1;
}

/** \brief Sets the marked bit for a reference.
 *
 * \param reference An object reference.
 * \param marked    The new value for the marked bit.
 */
static inline void ObjectTable_setMarked(U16 reference, bool marked) {
  NXT_ASSERT(!ObjectTable_isFree(reference), "Marking of a free object reference");
  ObjectTable_set(reference, MODIFY_BIT(ObjectTable_get(reference), MARK_BIT, marked));
}

/** \brief Marks an object reference.
 *
 * \param reference The object reference to mark.
 */
static inline void ObjectTable_mark(U16 reference) {
  ObjectTable_setMarked(reference, TRUE);
}

/** \brief Indicates, whether an object has been copied from Flash.
 *
 * \param reference An object reference.
 * \return TRUE, if the object has been moved from Flash. FALSE, if not.
 */
static inline bool ObjectTable_isRenewed(U16 reference) {
  return (ObjectTable_get(reference) >> RENEWED_BIT) & 1;
}

/** \brief Sets the renewed bit for a reference.
 *
 * \param reference An object reference.
 * \param renewed   The new value for the renewed bit.
 */
static inline void ObjectTable_setRenewed(U16 reference, bool renewed) {
  ObjectTable_set(reference, MODIFY_BIT(ObjectTable_get(reference), RENEWED_BIT, renewed));
}

/** \brief Indicates, whether an object is indexable.
 *
 * \param reference An object reference.
 * \return TRUE, if the object is indexable. FALSE, if not.
 */
static inline bool ObjectTable_isIndexable(U16 reference) {
  return (ObjectTable_get(reference) >> INDEXABLE_BIT) & 1;
}

/** \brief Sets the indexable bit for a reference.
 *
 * \param reference An object reference.
 * \param indexable The new value for the indexable bit.
 */
static inline void ObjectTable_setIndexable(U16 reference, bool indexable) {
  ObjectTable_set(reference, MODIFY_BIT(ObjectTable_get(reference), INDEXABLE_BIT, indexable));
}

/** \brief Indicates, whether an object consists of regular instance variables.
 *
 * \param reference An object reference.
 * \return TRUE, if the object references other objects. FALSE, if not.
 */
static inline bool ObjectTable_isReferencingInstances(U16 reference) {
  return (ObjectTable_get(reference) >> INSTANCES_BIT) & 1;
}

/** \brief Sets the references bit for a reference.
 *
 * \param reference     An object reference.
 * \param refInstances  The new value for the references bit.
 */
static inline void ObjectTable_setReferencingInstances(U16 reference, bool refInstances) {
  ObjectTable_set(reference, MODIFY_BIT(ObjectTable_get(reference), INSTANCES_BIT, refInstances));
}

#ifndef LINUX
/** \brief Fetches the memory address of an object.
 *
 * \param reference An object reference.
 * \return The address encoded in the object table entry.
 */
static inline void* ObjectTable_getAddress(U16 reference) {
  return (void*)(ObjectTable_get(reference) & 0x3FFFFF);
}

/** \brief Updates the memory address of an object.
 *
 * \param reference An object reference.
 * \param object    The new address for that object.
 */
static inline void ObjectTable_setAddress(U16 reference, void* object) {
  ObjectTable_set(reference, (int)object | (ObjectTable_get(reference) & ~0x3FFFFF));
}

#else
/** \brief Fetches the memory address of an object under Linux.
 *
 * \param reference An object reference.
 * \return The address encoded in the object table entry.
 */
static inline void* ObjectTable_getAddress(U16 reference) {
  U32 nxtAddress = ObjectTable_get(reference) & 0x3FFFFF;
  NXT_ASSERT(nxtAddress >= 0x100000 && nxtAddress < 0x240000, "Address out of range");
  if (nxtAddress < 0x200000) {
    return NXTFlash_getStart() + nxtAddress - 0x100000;
  } else {
    return NXTMemory_getStart() + nxtAddress - 0x200000;
  }
}

/** \brief Updates the memory address of an object under Linux.
 *
 * \param reference An object reference.
 * \param object    The new address for that object.
 */
static inline void ObjectTable_setAddress(U16 reference, void* object) {
  U32 nxtAddress;
  if (NXTFlash_getStart() <= (U8*)object && (U8*)object < NXTFlash_getEnd()) {
    nxtAddress = ((U8*)object - NXTFlash_getStart()) + 0x100000;
  } else {
    nxtAddress = ((U8*)object - NXTMemory_getStart()) + 0x200000;
  }
  NXT_ASSERT(nxtAddress >= 0x100000 && nxtAddress < 0x240000, "Address out of range");
  ObjectTable_set(reference, nxtAddress | (ObjectTable_get(reference) & ~0x3FFFFF));
}
#endif

/** \brief Checks, whether the given object lies in RAM under Linux.
 *
 * \param object An object reference.
 * \return TRUE, if the object lies in RAM. FALSE, if not.
 */
static inline bool ObjectTable_isRAMObject(U16 object) {
  return ObjectTable_isRealObject(object) &&
      NXTMemory_isRAMAddress(ObjectTable_getAddress(object));
}

/** \brief Checks, whether the given object lies in the Flash under Linux.
 *
 * \param object An object reference.
 * \return TRUE, if the object lies in the Flash. FALSE, if not.
 */
static inline bool ObjectTable_isROMObject(U16 object) {
  return ObjectTable_isRealObject(object) &&
      !NXTMemory_isRAMAddress(ObjectTable_getAddress(object));
}

/** \brief Finds a free table index.
 *
 * Finds a free table index or tries to allocate a new entry at the
 * bottom of the table.
 *
 * \return  The index of the table entry or 0, if the table is full
 *          and couldn't be resized.
 */
U16 ObjectTable_findFreeIndex(void);

/** \brief Searches the table index corresponding to a given address.
 *
 * Fails, if no entry with the provided address is found.
 *
 * \param address The address of the object to find.
 * \return The index, if one could be found.
 */
static inline U16 ObjectTable_find(void* address) {
  for (U32 reference = minReference; reference <= MAX_REFERENCE; reference += 2) {
    if (ObjectTable_getAddress(reference) == address) {
      return reference;
    }
  }
  return 0;
}

/** \brief Returns, whether the address in the object table is encoded.
 *
 * * The mark & sweep garbage collector replaces the class reference
 * of each object with the reference to the object itself in one of
 * its phases. In turn, it does not need the address bits in the
 * table entries. Thus, the class reference and size information
 * has to be encoded in the (currently) free address bits. This function
 * returns, whether the given table entry has encoded information or
 * not by looking at the last bit. Real addresses on the NXT have 0
 * here (16-bit alignment).
 *
 * \param reference The object reference to test.
 * \return TRUE, if the address contains encoded information.
 */
static inline bool ObjectTable_isEncoded(U16 reference) {
  return ObjectTable_get(reference) & 0x1;
}

/** \brief Encodes the class and size information in the object table.
 *
 * \param reference The object reference, where the address bits shall be
 *                  replaced with the class reference and size information.
 */
void ObjectTable_encodeClassAndSizeInfo(U16 reference);

/** \brief Decodes the class reference from the object table.
 *
 * \param reference The object reference, where the address bits contain the
 *                  class reference and size information.
 * \return The class reference.
 */
U16 ObjectTable_decodeClassInfo(U16 reference);

/** \brief Decodes the indexable bit from the object table.
 *
 * \param reference The object reference, where the address bits contain the
 *                  class reference and size information.
 * \return TRUE, if the object is indexable. FALSE, if not.
 * \deprecated Remove and use ObjectTable_isIndexable.
 */
//bool ObjectTable_decodeIsIndexableSize(U16 reference);

/** \brief Decodes the size for non-indexed objects from the object table.
 *
 * \param reference The object reference, where the address bits contain the
 *                  class reference and size information.
 * \return  The number of instance variables for objects which are not
 *          indexable. These would carry their own size within its data.
 */
U16 ObjectTable_decodeSizeForNonIndexableObject(U16 reference);

/** \brief Decodes the element size for indexed objects from the object table.
 *
 * \param reference The object reference, where the address bits contain the
 *                  class reference and size information.
 * \return The size of each element in the indexable object.
 * \deprecated Only needed for indicating, whether this object has regular instances.
 */
U16 ObjectTable_decodeElementSize(U16 reference);

#endif // _OBJECTTABLE_H_
