/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

/** \file primitives.h
 * \brief Support types and functions for primitives.
 * \author Martin Beck
 */

#ifndef _PRIMITIVES_H_
#define _PRIMITIVES_H_

#include "object.h"

/** \brief Helper structure for primitive stack frames.
 *
 * Helper structure which stores stack frame information like the receiver
 * and the argument count. Primitives can access the arguments (still lying
 * on the stack) more easier this way. Such a PrimitiveFrame is set up by
 * the interpreter before calling the primitive.
 */
struct PrimitiveFrame {

  /** \brief Reference to the receiver. */
  ObjectRef self;

  /** \brief Argument count for the primitive. */
  U16 argumentCount;
} __attribute__ ((packed));

/** \brief Function pointer to a primitive. */
typedef void(*Primitive)(PrimitiveFrame* frame);

/** \brief Maximum amount of primitives. */
#define PRIMITIVE_COUNT 83

/** \brief Maximum number of temporary objects within primitives. */
#define PRIMITIVE_TEMP_SET_SIZE 10

/** \brief An array of all possible primitives ordered by number. */
extern Primitive primitives[PRIMITIVE_COUNT];

/** \brief Set of temporaries within the current primitive. */
extern ObjectRef temporarySet[PRIMITIVE_TEMP_SET_SIZE];

/** \brief Size of the temporary set for the current primitive. */
extern U8 temporarySetSize;

/** \brief Fetches an argument.
 *
 * Fetches an argument from the current process' stack.
 *
 * \param frame The primitive frame meta data.
 * \param index The index of the argument.
 * \return The value of the argument.
 */
ObjectRef Primitive_argumentAt(PrimitiveFrame* frame, U16 index);

/** \brief Updates an argument.
 *
 * Updates an argument on the current process' stack.
 *
 * \param frame The primitive frame meta data.
 * \param index Index of the argument.
 * \param value New value for the argument.
 */
void Primitive_argumentAtPut(PrimitiveFrame* frame, U16 index, ObjectRef value);

/** \brief Clears the stack of the primitive.
 *
 * Clears the stack of the primitive by removing all arguments and the
 * receiver.
 *
 * \param frame The primitive frame meta data.
 */
void Primitive_clearStack(PrimitiveFrame* frame);

/** \brief Returns a value from the primitive.
 *
 * As primitive functions cannot return a value themselves, this helper
 * function clears the stack and pushes the given return value onto
 * it.
 *
 * \param frame The primitive frame meta data.
 * \param value Value to be returned to Smalltalk by the calling primitive.
 */
void Primitive_return(PrimitiveFrame* frame, ObjectRef value);

/** \brief Creates an instance of a class for intermediate primitive usage.
 *
 * If a primitive needs an intermediate instance, which it will not return,
 * this instance will not get cleaned up until the next full garbage
 * collection. Thus, this helper registers the new object for a cleanup
 * after the primitive has finished.
 *
 * \param klass         The klass to instanciate.
 * \param sizeRequested Number of data slots in new instance.
 * \return Reference to intermediate instance.
 */
ObjectRef Primitive_createIntermediateInstance(ObjectRef klass, U16 sizeRequested);

/** \brief Adds an object to the temporary set of the current primitive.
 *
 * \param object The object to add.
 */
void Primitive_addToTemporarySet(ObjectRef object);

/** \brief Clears the temporary set of the current primitive. */
void Primitive_emptyTemporarySet(void);


#endif // _PRIMITIVES_H_
