/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

/** \file scheduler.h
 * \brief Functions and primitives for the NXTProcess class.
 * \author Martin Beck
 */

#ifndef _PROCESS_H_
#define _PROCESS_H_

#include "object.h"
#include "frame.h"
#include "event.h"
#include "primitives.h"

/** \brief Type redefinition to indicate Process objects. */
typedef ObjectRef ProcessRef;

/** \brief Memory structure of process objects.
 *
 * A process is prepared to be inserted into a linked list and stores
 * the current frame, which it is executing. Furthermore, it has a
 * stack and the current top index of that stack. Nevertheless, this
 * data is only updated upon process switches, because it would be too
 * slow otherwise. The process may also contain a list of events, which
 * it waits for, if it is blocked.
 */
struct Process {

  /** \brief Class reference.*/
  ClassRef klass;

  /** \brief Link to next process, if in list. */
  ProcessRef nextLink;

  /** \brief Currently executed frame. */
  FrameRef currentFrame;

  /** \brief Process' stack.
   *
   * The stack of a process is used to store intermediate values like
   * return values and/or arguments (before they are copied into their
   * Frame object). It is represented as a simple array, which is filled
   * from the bottom to the top. The stack's initial size is 100 entries,
   * but it can grow if necessary.
   */
  ObjectRef stack;

  /** \brief Index of stack top. */
  SmallInteger stackTopIndex;

  /** \brief Linked list of event conditions this process is waiting for. */
  EventRef waitsFor;
} __attribute__ ((packed));
typedef struct Process Process;

/** \brief Default stack size. */
#define STACK_SIZE 500

/** \brief Resolve a Process reference.
 *
 * Retrieves a direct pointer to a Process object.
 *
 * \param processRef The reference to the process.
 * \return The direct pointer to the Process object in memory.
 */
static inline Process* Process_get(ProcessRef processRef) {
  return (Process*)Object_get(processRef);
}

/** \brief Resolve a Process reference for writing.
 *
 * Retrieves a direct pointer to a Process object ensuring it lies
 * within the writeable RAM region.
 *
 * \param processRef The reference to the process.
 * \return The direct pointer to the Process object in RAM.
 */
static inline Process* Process_getForWrite(ProcessRef processRef) {
  return (Process*)Object_getForWrite(processRef);
}

/** \brief Instanciates a new process with a default-sized stack.
 *
 * \return A reference to the new Process object.
 */
ProcessRef Process_new(void);

/** \brief Checks for the events the process is waiting for.
 *
 * \param self The process.
 */
bool Process_checkEvents(const Process* self);

/** \name Getters
 * @{
 */

/** \brief Returns the stack's top index as a signed integer.
 *
 * \param self The process.
 * \return The index of process' stack's top element.
 */
static inline S16 Process_getStackTopIndex(Process* self) {
  return SmallInteger_fromReference(self->stackTopIndex);
}
/** @} */

/** \name Setters
 * @{
 */

/** \brief Sets the current frame.
 *
 * Sets the current frame and does NO reference counting. This is not
 * needed, because the interpreter always holds a reference to the
 * current frame implicitly. If the current frame is changed from
 * outside the process, that happens from the Smalltalk side and thus
 * the correct reference counting is done. I.e., it is not required
 * to use this function, as the currentFrame can be set directly without
 * any further actions.
 *
 * \param self  The process.
 * \param frame The new frame.
 */
static inline void Process_setCurrentFrame(Process* self, FrameRef frame) {
  self->currentFrame = frame;
}

/** \brief Sets a new stack and does reference counting accordingly.
 *
 * \param self  The process.
 * \param stack The new stack.
 */
static inline void Process_setStack(Process* self, ObjectRef stack) {
  Object_incReferenceCount(stack);
  Object_decReferenceCount(self->stack);
  self->stack = stack;
}

/** \brief Sets the stack top index.
 *
 * \param self          The process.
 * \param stackTopIndex The new stack top index as signed integer.
 */
static inline void Process_setStackTopIndex(Process* self, S16 stackTopIndex) {
  self->stackTopIndex = SmallInteger_toReference(stackTopIndex);
}
/** @} */

/** \name Primitives
 * @{
 */

/** \brief Returns from the specified context in the sender chain. */
void Process_primReturnFrom(PrimitiveFrame* frame);
/** @} */

#endif // _PROCESS_H_
