/***************************************************************************
 * Copyright (c) 2008 by Martin Beck                                       *
 * martin.beck@hpi.uni-potsdam.de                                          *
 *                                                                         *
 * Permission is hereby granted, free of charge, to any person             *
 * obtaining a copy of this software and associated documentation          *
 * files (the "Software"), to deal in the Software without                 *
 * restriction, including without limitation the rights to use,            *
 * copy, modify, merge, publish, distribute, sublicense, and/or sell       *
 * copies of the Software, and to permit persons to whom the               *
 * Software is furnished to do so, subject to the following                *
 * conditions:                                                             *
 *                                                                         *
 * The above copyright notice and this permission notice shall be          *
 * included in all copies or substantial portions of the Software.         *
 *                                                                         *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,         *
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES         *
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                *
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT             *
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,            *
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING            *
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR           *
 * OTHER DEALINGS IN THE SOFTWARE.                                         *
 ***************************************************************************/

/** \file scheduler.h
 * \brief Functions and primitives for the NXTProcessorScheduler class.
 * \author Martin Beck
 */

#ifndef _SCHEDULER_H_
#define _SCHEDULER_H_

#include "process.h"

/** \brief Type redefinition to indicate a reference to the Processor object. */
typedef ObjectRef SchedulerRef;

/** \brief Memory structure of the scheduler.
 *
 * The scheduler class inherits from NXTLinkedList. Hence, its instances
 * have a firstLink and a lastLink instance variable. Afterwards, the
 * additional variables for the active and scheduling processes and a
 * waiting list follow. */
struct Scheduler {

  /** \brief Class reference. */
  ClassRef klass;

  /** \brief The first process in the scheduling chain. */
  ProcessRef firstLink;

  /** \brief The last process in the scheduling chain. */
  ProcessRef lastLink;

  /** \brief The active process excluding the scheduler process. */
  ProcessRef activeProcess;

  /** \brief A linked list of processes waiting for an external event. */
  ProcessRef waitingProcesses;

  /** \brief A reference to the scheduler process. */
  ProcessRef schedulerProcess;
} __attribute__((packed));
typedef struct Scheduler Scheduler;

/** \brief Indicates, whether the scheduler process is currently running. */
extern bool schedulerRunning;

/** \brief Indicates, whether the scheduling is currently enabled. */
extern bool schedulingEnabled;

/** \brief Resolves a reference to a Scheduler.
 *
 * Fetches a direct pointer to a Scheduler from the object table.
 *
 * \param scheduler A reference to a scheduler object.
 * \return Direct pointer to the Scheduler.
 */
static inline const Scheduler* Scheduler_get(SchedulerRef scheduler) {
  return (Scheduler*)Object_get(scheduler);
}

/** \brief Resolves a reference to a Scheduler for writing.
 *
 * Fetches a direct pointer to a Scheduler from the object table and
 * ensures, it lies in the writeable RAM region.
 *
 * \param scheduler A reference to a scheduler object.
 * \return Direct pointer to the Scheduler within RAM.
 */
static inline Scheduler* Scheduler_getForWrite(SchedulerRef scheduler) {
  return (Scheduler*)Object_getForWrite(scheduler);
}

/** \brief Checks, whether waiting process can be awakened.
 *
 * Iterates through the waiting processes list and checks
 * every process for its event conditions. If the event
 * occured, the process is awakened and the iteration
 * exited.
 *
 * \param self The scheduler.
 */
void Scheduler_checkWaitingProcesses(Scheduler* self);

/** \name Setters
 * @{
 */

/** \brief Sets the active process using reference counting.
 *
 * \param self          The scheduler.
 * \param activeProcess The new active process.
 */
static inline void Scheduler_setActiveProcess(Scheduler* self, ProcessRef activeProcess) {
  Object_incReferenceCount(activeProcess);
  Object_decReferenceCount(self->activeProcess);
  self->activeProcess = activeProcess;
}

/** \brief Sets the scheduler process using reference counting.
 *
 * \param self              The scheduler.
 * \param schedulerProcess  The new scheduler process.
 */
static inline void Scheduler_setSchedulerProcess(Scheduler* self, ProcessRef schedulerProcess) {
  Object_incReferenceCount(schedulerProcess);
  Object_decReferenceCount(self->schedulerProcess);
  self->schedulerProcess = schedulerProcess;
}
/** @} */

/** \name Primitives
 * @{
 */

/** \brief Schedule to a new active process. */
void Scheduler_primScheduleTo(PrimitiveFrame* frame);

/** \brief Enable or disable scheduling. */
void Scheduler_primSchedulingEnabled(PrimitiveFrame* frame);

/** \brief Yield to the scheduler process. */
void Scheduler_primYield(PrimitiveFrame* frame);

/** \brief Check waiting process for their event conditions. */
void Scheduler_primCheckWaitingProcesses(PrimitiveFrame* frame);
/** @} */

#endif // #_SCHEDULER_H_
